package dischoco;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.Date;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JToolBar;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.Timer;
import javax.swing.UIManager;
import javax.swing.UIManager.LookAndFeelInfo;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;

import benchmark.graphcoloring.URGCGenerator;
import benchmark.graphcoloring.gui.GCGeneratorPanel;
import benchmark.meetingscheduling.URMeetingScheduling;
import benchmark.nqueen.NQueenGenerator;
import benchmark.random.RBCDisCSPGenerator;
import benchmark.random.RBDisCSPGenerator;
import benchmark.rmsp.RandomMSP;
import benchmark.sensornetwork.SensorDCSPGenerator;
import gui.Filter;
import gui.GuiConsole;
import gui.GuiSimulator;
import gui.LabelAccessory;
import gui.debug.DebugPanel;
import gui.experiments.ConfigurationPanel;
import gui.generators.GeneratorsPanel;
import gui.results.ResultsPanel;
import gui.simulation.SimulationOnBenchmarks;
import gui.simulation.SimulationsPanel;
import gui.visualizing.VisualizingPanel;
import util.Commands;

/**
 * 
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">Redoaune Ezzahir</a>
 * and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr"> Mohamed Wahbi</a>
 * 
 */
public class DisChoco extends JFrame {
	
	public static final String authors = "<html>" +
	"<body>" +
	"<center>" +
	"for Education Copyright (c) 2005-2011 by:" +
	"<a href=\"mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr\">" +
	"Redoaune Ezzahir</a> and " +
	"<a href=\"mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr\"> " +
	"Mohamed Wahbi</a>" +
	"</center>" +
	"</body>" +
	"</html>";
	
	private static final long serialVersionUID = -2955136742478294106L;
	private boolean consoleActive = true;
	private JPanel mainPanel = null;
	private JMenuBar myMenuBar = null;
	private JMenu fileMenu = null;
	private JMenu generatorsMenu = null;
	private JMenu experimentsMenu = null;
	private JMenu resultsMenu = null;
	private JMenu themesMenu = null; // themes
	private JMenu simulationMenu = null;
	private JMenu debugMenu = null;
	private JMenu helpMenu = null;
	private JMenuItem exitMenuItem = null;
	private JMenuItem aboutMenuItem = null;
	private JMenuItem randomDisCSPMenuItem = null;
	private JMenuItem complexRandomDisCSPMenuItem = null;
	private JMenuItem loadProblemMenuItem = null;
	private JMenuItem loadResultsMenuItem = null;
	private JMenuItem meetingSchedulingMenuItem = null;
	private JMenuItem randomMSPMenuItem = null;
	private JMenuItem sensorNetworksMenuItem = null;
	private JMenuItem graphColoringMenuItem = null;
	private JMenuItem complexGraphColoringMenuItem = null;
	private JMenuItem nQueenMenuItem = null;
	private JMenuItem randomDCOPMenuItem = null;
	private JMenuItem generateProblemsMenuItem = null;
	private JMenuItem configurationMenuItem = null;
	private JMenuItem loadConfigurationFileMenuItem = null;
	private JMenuItem plotResultsMenuItem = null;
	private JMenuItem mergeResultsMenuItem = null;
	private JMenuItem debugAlgorithmMenuItem = null;
	private JMenuItem visualizeMessageMenuItem = null;
	private JMenuItem visualizeAgentsLoadMenuItem = null;
	private JDialog aboutDialog = null;
	private JPanel aboutContentPane = null;
	private JLabel aboutVersionLabel = null;
	private GeneratorsPanel generatorsPanel = null;
	private ResultsPanel resultsPanel = null;
	private VisualizingPanel visualizingPanel = null;
	private DebugPanel debugPanel = null;
	private SimulationsPanel simulationPanel = null;
	private ConfigurationPanel configurationPanel = null;
	// private GuiConsole guiConsole = null;
	private JScrollPane consoleScrollPane = null;
	private JPanel emptyPanel = null;
	private JPanel southPanel = null;
	private JPanel northPanel = null;//Toolkit.class.getCanonicalName();//
	public static String size32 = "32x32";
	public static String size24 = "24x24";
	public static String IconsSize = size24 + File.separator;
	public static String IconsPath = File.separator + IconsSize ;
	
	public final static String EMPTY = "empty JPanel";
	public final static String GENERTORSPANEL = "generators JPanel";
	public final static String CONFIGPANEL = "configuration JPanel";
	public final static String RESULTS_PANEL = "results JPanel";
	public final static String VISUALIZING_PANEL = "visualizing JPanel";
	public final static String DEBUG_PANEL = "debug JPanel";
	public final static String SIMULATION_PANEL = "simulation JPanel";

	private JToolBar mainToolBar = null;
	private JButton runButton = null;
	private JButton pauseButton = null;
	private JButton stopButton = null;
	private JButton helpButton = null;
	private JButton activeConsoleButton = null;
	private JButton plotButton;
	private JButton mergeButton;
	private JFileChooser fileChooser = null;
	private GuiSimulator simulator = null;
	private boolean isRunning = false;
	private LookAndFeelInfo[] looks = null;
	private ImageIcon logoDischoco = null;
	private JLabel logoIconLabel = null;
	private GuiConsole console;
	private JLabel authorsLabel = null;
	private JLabel webSiteLabel = null;
	private JMenu exportAsMenuItem;
	public static File filePath = null;
	
	/**
	 * 
	 */
	public DisChoco() {
		super();
		filePath = new File(".");
		logoDischoco = new ImageIcon(getClass().getResource(IconsPath + "logo.png"));
		setJMenuBar(getMyMenuBar());
		initialize();
		getContentPane().add(getMainPanel(), BorderLayout.CENTER);
		getContentPane().add(getMainToolBar(), BorderLayout.NORTH);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setTitle("DisChoco: Distributed Constraint Reasoning Solver");
		setMinimumSize(new Dimension(1024, 720));
		setMaximumSize(new Dimension(1024, 720));
		setPreferredSize(new Dimension(1024, 720));
		Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
		// Determine the new location of the window
		int w = getSize().width;
		int h = getSize().height;
		int x = (dim.width - w) / 2;
		int y = (dim.height - h) / 2;
		// Move the window
		setLocation(x, y);
	}
	
	/**
	 * 
	 * @return
	 */
	public JFrame getMainFrame(){
		return this;
	}
	
	/**
	 * This method initializes resultsMenu
	 * 
	 * @return javax.swing.JMenu
	 */
	private JMenu getThemesMenu() {
		if (themesMenu == null) {
			themesMenu = new JMenu();
			themesMenu.setText("Themes");
			looks = UIManager.getInstalledLookAndFeels();
			JMenuItem[] themes = new JMenuItem[looks.length];
			for (int t = 0 ; t < themes.length ; t++) {
				themes[t] = new JMenuItem();
				final String className = looks[t].getClassName();
				String name = looks[t].getName();
				themes[t].setText(name);
				themes[t].addActionListener(new ActionListener() {
					public void actionPerformed(ActionEvent e) {
						try {
							UIManager.setLookAndFeel(className);
							SwingUtilities.updateComponentTreeUI(getMainFrame());
						} catch (Exception exception) {
							JOptionPane.showMessageDialog(getMainFrame(), "Can't change look and feel", "Invalid PLAF", JOptionPane.ERROR_MESSAGE);
						}
					}
				});
				themesMenu.add(themes[t]);
			}
			themesMenu.addSeparator();
			JMenuItem iconsSize = new JMenuItem();
			iconsSize.setText("Icons size");
			themesMenu.add(iconsSize);
		}
		return themesMenu;
	}

	private void initialize() {
		getGeneratorsPanel();
		getConfigurationPanel();
		getResultsPanel();
		getVisualizingPanel();
		getSimulationPanel();
		getDebugPanel();
	}

	private JPanel getEmptyPanel() {
		if (emptyPanel == null) {
			emptyPanel = new JPanel();
			emptyPanel.setLayout(new BorderLayout());
			emptyPanel.setBackground(Color.lightGray);
			emptyPanel.setPreferredSize(new Dimension(1024, 520));
		}
		return emptyPanel;
	}

	private JPanel getMainPanel() {
		if (mainPanel == null) {
			mainPanel = new JPanel(new BorderLayout());
			mainPanel.add(getNorthPanel(), BorderLayout.CENTER);
			mainPanel.add(getSouthPanel(1024, 200), BorderLayout.SOUTH);
		}
		return mainPanel;
	}

	/**
	 * This method initializes southPanel
	 * 
	 * @param height
	 * @param width
	 * 
	 * @return javax.swing.JPanel
	 */
	private JPanel getNorthPanel() {
		if (northPanel == null) {
			northPanel = new JPanel(new CardLayout());
			northPanel.add(getEmptyPanel(), EMPTY);
			northPanel.add(generatorsPanel, GENERTORSPANEL);
			northPanel.add(configurationPanel, CONFIGPANEL);
			northPanel.add(resultsPanel, RESULTS_PANEL);
			northPanel.add(visualizingPanel, VISUALIZING_PANEL);
			northPanel.add(debugPanel, DEBUG_PANEL);
			northPanel.add(simulationPanel, SIMULATION_PANEL);
		}
		return northPanel;
	}

	/**
	 * This method initializes southPanel
	 * 
	 * @param height
	 * @param width
	 * 
	 * @return javax.swing.JPanel
	 */
	private JPanel getSouthPanel(int width, int height) {
		if (southPanel == null) {
			southPanel = new JPanel();
			southPanel.setLayout(new BorderLayout());
			southPanel.setBackground(Color.lightGray);
			southPanel.add(getConsoleScrollPane(width, height), BorderLayout.CENTER);
			southPanel.add(new JStatusBar(width), BorderLayout.SOUTH);
		}
		return southPanel;
	}

	/**
	 * This method initializes consoleScrollPane
	 * 
	 * @param height
	 * @param width
	 * 
	 * @return javax.swing.JScrollPane
	 */
	private JScrollPane getConsoleScrollPane(int width, int height) {
		if (consoleScrollPane == null) {
			consoleScrollPane = new JScrollPane(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
			consoleScrollPane.setViewportView(getConsole());
			consoleScrollPane.setPreferredSize(new Dimension(width, height));
		}
		return consoleScrollPane;
	}

	/**
	 * @return
	 */
	public GuiConsole getConsole() {
		if (console == null) {
			console = new GuiConsole(this.consoleActive);
		}
		return console;
	}

	/**
	 * This method initializes generateProblemsMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getGenerateProblemsMenuItem() {
		if (generateProblemsMenuItem == null) {
			generateProblemsMenuItem = new JMenuItem();
			generateProblemsMenuItem.setText("Generate Problems");
			generateProblemsMenuItem.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), GENERTORSPANEL);
				}
			});
		}
		return generateProblemsMenuItem;
	}

	/**
	 * This method initializes configurationMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getConfigurationMenuItem() {
		if (configurationMenuItem == null) {
			configurationMenuItem = new JMenuItem();
			configurationMenuItem.setText("Configure experimentation");
			configurationMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					CardLayout cards = (CardLayout) (getNorthPanel().getLayout());
					cards.show(getNorthPanel(), CONFIGPANEL);
				}
			});
		}
		return configurationMenuItem;
	}

	/**
	 * This method initializes graphColoringMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getGraphColoringMenuItem() {
		if (graphColoringMenuItem == null) {
			graphColoringMenuItem = new JMenuItem();
			graphColoringMenuItem.setText("Graph Coloring");
			graphColoringMenuItem.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), GENERTORSPANEL);
					GCGeneratorPanel gcPanel = generatorsPanel.getGraphColoringPanel();
					gcPanel.setFormat(Commands.SIMPLE);
					generatorsPanel.setGenerator(GeneratorsPanel.graphColoring);
				}
			});
		}
		return graphColoringMenuItem;
	}

	/**
	 * This method initializes complexGraphColoringMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getComplexGraphColoringMenuItem() {
		if (complexGraphColoringMenuItem == null) {
			complexGraphColoringMenuItem = new JMenuItem();
			complexGraphColoringMenuItem.setText("Complex Graph Coloring");
			complexGraphColoringMenuItem.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), GENERTORSPANEL);
					generatorsPanel.setGenerator(GeneratorsPanel.graphColoring);
					GCGeneratorPanel gcPanel = generatorsPanel.getGraphColoringPanel();
					gcPanel.setFormat(Commands.COMPLEX);
				}
			});
		}
		return complexGraphColoringMenuItem;
	}

	/**
	 * This method initializes jMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getRandomDisCSPMenuItem() {
		if (randomDisCSPMenuItem == null) {
			randomDisCSPMenuItem = new JMenuItem();
			randomDisCSPMenuItem.setText("Random DisCSP");
			randomDisCSPMenuItem.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), GENERTORSPANEL);
					generatorsPanel.setGenerator(GeneratorsPanel.randomDisCSP);
				}
			});
		}
		return randomDisCSPMenuItem;
	}

	/**
	 * This method initializes meetingSchedulingMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getMeetingSchedulingMenuItem() {
		if (meetingSchedulingMenuItem == null) {
			meetingSchedulingMenuItem = new JMenuItem();
			meetingSchedulingMenuItem.setText("Meeting Scheduling");
			meetingSchedulingMenuItem.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), GENERTORSPANEL);
					generatorsPanel.setGenerator(GeneratorsPanel.meetingScheduling);
				}
			});
		}
		return meetingSchedulingMenuItem;
	}

	/**
	 * This method initializes meetingSchedulingMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getRandomMSPMenuItem() {
		if (randomMSPMenuItem == null) {
			randomMSPMenuItem = new JMenuItem();
			randomMSPMenuItem.setText("Random Meeting Scheduling");
			randomMSPMenuItem.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), GENERTORSPANEL);
					generatorsPanel.setGenerator(GeneratorsPanel.randomMSP);
				}
			});
		}
		return randomMSPMenuItem;
	}

	/**
	 * This method initializes sensorNetworksMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getSensorNetworksMenuItem() {
		if (sensorNetworksMenuItem == null) {
			sensorNetworksMenuItem = new JMenuItem();
			sensorNetworksMenuItem.setText("Sensor Networks");
			sensorNetworksMenuItem.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), GENERTORSPANEL);
					generatorsPanel.setGenerator(GeneratorsPanel.sensorNetworks);
				}
			});
		}
		return sensorNetworksMenuItem;
	}

	/**
	 * This method initializes nQueenMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getNQueenMenuItem() {
		if (nQueenMenuItem == null) {
			nQueenMenuItem = new JMenuItem();
			nQueenMenuItem.setText("N Queen");
			nQueenMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), GENERTORSPANEL);
					generatorsPanel.setGenerator(GeneratorsPanel.nQueens);
				}
			});
		}
		return nQueenMenuItem;
	}

	/**
	 * This method initializes jMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getComplexRandomDisCSPMenuItem() {
		if (complexRandomDisCSPMenuItem == null) {
			complexRandomDisCSPMenuItem = new JMenuItem();
			complexRandomDisCSPMenuItem.setText("Complex Random DisCSP");
			complexRandomDisCSPMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), GENERTORSPANEL);
					generatorsPanel.setGenerator(GeneratorsPanel.randomComplexDisCSP);
				}
			});
		}
		return complexRandomDisCSPMenuItem;
	}

	/**
	 * This method initializes randomDCOPMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getRandomDCOPMenuItem() {
		if (randomDCOPMenuItem == null) {
			randomDCOPMenuItem = new JMenuItem();
			randomDCOPMenuItem.setText("Random DCOP");
			randomDCOPMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					printlnWarning("random DCOP generator is not yet implemented");
				}
			});
		}
		return randomDCOPMenuItem;
	}

	/**
	 * This method initializes loadConfigurationFileMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getLoadConfigurationFileMenuItem() {
		if (loadConfigurationFileMenuItem == null) {
			loadConfigurationFileMenuItem = new JMenuItem();
			loadConfigurationFileMenuItem.setText("Load experiments file");
			loadConfigurationFileMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (loadExperimentsFile()) {
						runButton.setEnabled(true);
						pauseButton.setEnabled(false);
						stopButton.setEnabled(false);
					} else if(filePath == null) {
						runButton.setEnabled(false);
						pauseButton.setEnabled(false);
						stopButton.setEnabled(false);
					}
				}
			});
		}
		return loadConfigurationFileMenuItem;
	}

	/**
	 * 
	 */
	public boolean loadExperimentsFile() {
		String description = "experiments file (XML)";
		String[] extension = new String[] {"xml"};
		Filter filter = new Filter(extension, description);
		fileChooser = new JFileChooser(filePath.getPath());
		fileChooser.setFileFilter(filter);
//		fileChooser.setFileView(new JavaFileView());
		fileChooser.setPreferredSize(new Dimension(660, 440));
		int returnVal = fileChooser.showOpenDialog(getMainFrame());
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			filePath = fileChooser.getSelectedFile();
			println("The file \"" + filePath + "\" was successfelly loaded!");
			Document document = null;
			SAXBuilder sxb = new SAXBuilder();
			try {
				document = sxb.build(filePath);
			} catch (Exception ex) { ex.printStackTrace(); }
			Element root = document.getRootElement();
			String type = root.getAttributeValue(Commands.type);
			if (type.equals(Commands.RandomTKC) || type.equals(Commands.RandomPKC)) {
				this.simulator = new RBDisCSPGenerator();
			} else if (type.equals(Commands.GraphColoring)) {
				this.simulator = new URGCGenerator();
			} else if (type.equals(Commands.NQueen)) {
				this.simulator = new NQueenGenerator();
			} else if (type.equals(Commands.ComplexRandomTKC)) {
				this.simulator = new RBCDisCSPGenerator();
			} else if (type.equals(Commands.SensorNetworks)) {
				this.simulator = new SensorDCSPGenerator();
			} else if (type.equals(Commands.MeetingScheduling)) {
				this.simulator = new URMeetingScheduling();
			} else if (type.equals(Commands.RMSP)) {
				this.simulator = new RandomMSP();
			}
			return true;
		} else {
			printlnWarning("No file loaded!");
			return false;
		}
	}

	/**
	 * @return javax.swing.JButton
	 */
	@SuppressWarnings("rawtypes")
	public JButton getRunButton() {//filePath
		if (runButton == null) {
			runButton = new JButton(new ImageIcon(getClass().getResource(IconsPath + "play.png")));
			runButton.setEnabled(false);
			runButton.setToolTipText("Start Experiment Run execution");
			runButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					runButton.setEnabled(false);
					pauseButton.setEnabled(true);
					stopButton.setEnabled(true);
					SwingWorker worker = new SwingWorker() {
						protected Object doInBackground() throws Exception {
							startRun();
							return null;
						}
					};
					worker.execute();
				}
			});
		}
		return runButton;
	}

	/**
	 * 
	 */
	private void startRun() {
		if(filePath.isDirectory()){
			JOptionPane.showMessageDialog(this, 
					"Please load the experiments file", 
					"No file experiments loaded", 
					JOptionPane.ERROR_MESSAGE);
			stop();
			return;
		}
		if(isRunning){
			println("Resume execution of experiments...");
			simulator.resume();
			stop();
			println("Congratulation!! your experiments were done successfully.");  
		} else {
			this.isRunning = true;
			println("Starting execution of experiments...");
			println("experiments: " + filePath);
			simulator.lunch();
			stop();
			println("Congratulation!! your experiments were done successfully.");  
		}
	}

	/**
	 * 
	 */
	private void stop() {
		runButton.setEnabled(false);
		pauseButton.setEnabled(false);
		stopButton.setEnabled(false);
		isRunning = false;
		simulator.stop();
//		filePath = null;
	}

	/**
	 * 
	 * @return
	 */
	public JButton getPauseButton() {
		if (pauseButton == null) {
			pauseButton = new JButton(new ImageIcon(getClass().getResource(IconsPath + "pause.png")));
			pauseButton.setEnabled(false);
			pauseButton.setToolTipText("Pause execution after finishing current problem");
			pauseButton.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (isRunning) {
						runButton.setEnabled(true);
						pauseButton.setEnabled(false);
						println("Pause execution of experiments...");
						simulator.pause();
					}
				}
			});
		}
		return pauseButton;
	}

	/**
	 * 
	 * @return
	 */
	public JButton getStopButton() {
		if (stopButton == null) {
			stopButton = new JButton(new ImageIcon(getClass().getResource(IconsPath + "stop.png")));
			stopButton.setEnabled(false);
			stopButton.setToolTipText("Stop Experiment Run execution");
			stopButton.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (isRunning) {
						stop();
						println("Stop running experimentations");
					}
				}
			});
		}
		return stopButton;
	}

	/**
	 * 
	 * @return
	 */
	public JButton getHelpButton() {
		if (helpButton == null) {
			helpButton = new JButton(new ImageIcon(getClass().getResource(IconsPath + "help.png")));
			helpButton.setToolTipText("Display user manual");
			helpButton.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(ActionEvent e) {
				// TODO
					printlnWarning("Manuel not yet implemented");
				}
			});
			helpButton.setMargin(new Insets(0, 0, 0, 0));
			helpButton.setVerticalTextPosition(JButton.BOTTOM);
			helpButton.setHorizontalTextPosition(JButton.CENTER);
		}
		return helpButton;
	}

	/**
	 * This method initializes plotResultsMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getPlotResultsMenuItem() {
		if (plotResultsMenuItem == null) {
			plotResultsMenuItem = new JMenuItem();
			plotResultsMenuItem.setText("Plot graphs");
			plotResultsMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), RESULTS_PANEL);
				}
			});
		}
		return plotResultsMenuItem;
	}

	/**
	 * This method initializes mergeResultsMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getMergeResultsMenuItem() {
		if (mergeResultsMenuItem == null) {
			mergeResultsMenuItem = new JMenuItem();
			mergeResultsMenuItem.setText("Merge results");
			mergeResultsMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), RESULTS_PANEL);
				}
			});
		}
		return mergeResultsMenuItem;
	}

	/**
	 * This method initializes debugMenu
	 * 
	 * @return javax.swing.JMenu
	 */
	private JMenu getSimulationMenu() {
		if (simulationMenu == null) {
			simulationMenu = new JMenu();
			simulationMenu.setText("Simulation");
			simulationMenu.add(getVisualizeMessageMenuItem());
			simulationMenu.add(getVisualizeAgentsLoadMenuItem());
		}
		return simulationMenu;
	}

	/**
	 * This method initializes visualizeMessageMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getVisualizeMessageMenuItem() {//TODO
		if (visualizeMessageMenuItem == null) {
			visualizeMessageMenuItem = new JMenuItem();
			visualizeMessageMenuItem.setText("Visualize exchanged messages");
//			visualizeMessageMenuItem.addActionListener(new ActionListener(){
//				public void actionPerformed(ActionEvent e) {
//					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
//					cards.show(getNorthPanel(), SIMULATION_PANEL);
//				}
//			});
			visualizeMessageMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					String description = "files (xml, bz2)";
					String[] extensions = new String[] {"xml", "bz2"};
					Filter filter = new Filter(extensions, description);
					fileChooser = new JFileChooser(filePath.getPath());
					fileChooser.setFileFilter(filter);
//					fileChooser.setFileView(new JavaFileView());
					fileChooser.setAccessory(new LabelAccessory(fileChooser));
					fileChooser.setPreferredSize(new Dimension(660, 440));
					int returnVal = fileChooser.showOpenDialog(getMainFrame());
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						CardLayout cards = (CardLayout) getNorthPanel().getLayout();
						cards.show(getNorthPanel(), SIMULATION_PANEL);
						filePath = fileChooser.getSelectedFile();
						println("The file \"" + filePath + "\" is successfelly loaded!");
						simulationPanel.setProblem(filePath, SimulationOnBenchmarks.MessagesVisualizing);
					} else {
						printlnWarning("No file loaded!");
					}
				}
			});
		}
		return visualizeMessageMenuItem;
	}
	/**
	 * This method initializes debugAlgorithmMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getVisualizeAgentsLoadMenuItem() {//TODO
		if (visualizeAgentsLoadMenuItem == null) {
			visualizeAgentsLoadMenuItem = new JMenuItem();
			visualizeAgentsLoadMenuItem.setText("Visualize agents load (CCs)");
//			visualizeAgentsLoadMenuItem.addActionListener(new ActionListener(){
//				public void actionPerformed(ActionEvent e) {
//					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
//					cards.show(getNorthPanel(), SIMULATION_PANEL);
//				}
//			});
			visualizeAgentsLoadMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					String description = "files (xml, bz2)";
					String[] extensions = new String[] {"xml", "bz2"};
					Filter filter = new Filter(extensions, description);
					fileChooser = new JFileChooser(filePath.getPath());
					fileChooser.setFileFilter(filter);
//					fileChooser.setFileView(new JavaFileView());
					fileChooser.setAccessory(new LabelAccessory(fileChooser));
					fileChooser.setPreferredSize(new Dimension(660, 440));
					int returnVal = fileChooser.showOpenDialog(getMainFrame());
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						CardLayout cards = (CardLayout) getNorthPanel().getLayout();
						cards.show(getNorthPanel(), SIMULATION_PANEL);
						filePath = fileChooser.getSelectedFile();
						println("The file \"" + filePath + "\" is successfelly loaded!");
						simulationPanel.setProblem(filePath, SimulationOnBenchmarks.AgentsLoad);
					} else {
						printlnWarning("No file loaded!");
					}
				}
			});
		}
		return visualizeAgentsLoadMenuItem;
	}
	
	/**
	 * This method initializes debugMenu
	 * 
	 * @return javax.swing.JMenu
	 */
	private JMenu getDebugMenu() {
		if (debugMenu == null) {
			debugMenu = new JMenu();
			debugMenu.setText("Debug");
			debugMenu.add(getDebugAlgorithmMenuItem());
		}
		return debugMenu;
	}

	/**
	 * This method initializes debugAlgorithmMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getDebugAlgorithmMenuItem() {
		if (debugAlgorithmMenuItem == null) {
			debugAlgorithmMenuItem = new JMenuItem();
			debugAlgorithmMenuItem.setText("Debug an algorithm");
			debugAlgorithmMenuItem.addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), DEBUG_PANEL);
				}
			});
		}
		return debugAlgorithmMenuItem;
	}

	/**
	 * This method initializes jJToolBarBar
	 * 
	 * @return javax.swing.JToolBar
	 */
	private JToolBar getMainToolBar() {
		if (mainToolBar == null) {
			mainToolBar = new JToolBar();
			mainToolBar.setFloatable(false);
			mainToolBar.setRequestFocusEnabled(false);
			mainToolBar.setMargin(new Insets(1, 1, 1, 1));
			mainToolBar.putClientProperty("JToolBar.isRollover", Boolean.TRUE);
			mainToolBar.add(getRunButton(), null);
			mainToolBar.add(getPauseButton(), null);
			mainToolBar.add(getStopButton(), null);
			mainToolBar.addSeparator();
			mainToolBar.add(getActiveConsoleButton(), null);
			mainToolBar.addSeparator();
			mainToolBar.add(getPlotButton(), null);
			mainToolBar.add(getMergeButton(), null);
			mainToolBar.addSeparator();
			mainToolBar.add(getHelpButton(), null);
		}
		return mainToolBar;
	}

	/**
	 * 
	 * @return
	 */
	public JButton getActiveConsoleButton() {
		if (activeConsoleButton == null) {
			activeConsoleButton = new JButton(new ImageIcon(getClass().getResource(IconsPath + "lock.png")));
			activeConsoleButton.setToolTipText("Enable/Disable the Gui console");
			activeConsoleButton.addActionListener(new java.awt.event.ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (consoleActive) {
						activeConsoleButton.setIcon(new ImageIcon(getClass().getResource(IconsPath + "lock.png")));
					} else {
						activeConsoleButton.setIcon(new ImageIcon(getClass().getResource(IconsPath + "lock_off.png")));
					}
					getConsole().setActive(consoleActive);
					consoleActive = !consoleActive;
				}
			});
			activeConsoleButton.setMargin(new Insets(0, 0, 0, 0));
			activeConsoleButton.setVerticalTextPosition(JButton.BOTTOM);
			activeConsoleButton.setHorizontalTextPosition(JButton.CENTER);
		}
		return activeConsoleButton;
	}

	/**
	 * 
	 * @return
	 */
	public JButton getPlotButton() {
		if (plotButton == null) {
			plotButton = new JButton(new ImageIcon(getClass().getResource(IconsPath + "chart.png")));
			plotButton.setToolTipText("Plot results");
			plotButton.setMargin(new Insets(0, 0, 0, 0));
			plotButton.setVerticalTextPosition(JButton.BOTTOM);
			plotButton.setHorizontalTextPosition(JButton.CENTER);
			plotButton.setEnabled(false);
		}
		return plotButton;
	}

	/**
	 * 
	 * @return
	 */
	public JButton getMergeButton() {
		if (mergeButton == null) {
			mergeButton = new JButton(new ImageIcon(getClass().getResource(IconsPath + "calculator.png")));
			mergeButton.setToolTipText("Merge results");
			mergeButton.setMargin(new Insets(0, 0, 0, 0));
			mergeButton.setVerticalTextPosition(JButton.BOTTOM);
			mergeButton.setHorizontalTextPosition(JButton.CENTER);
			mergeButton.setEnabled(false);
		}
		return mergeButton;
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		for (UIManager.LookAndFeelInfo laf : UIManager.getInstalledLookAndFeels()) {
			if ("Mac OS X".equals(laf.getName())) {
				try {
					UIManager.setLookAndFeel(laf.getClassName());
				} catch (Exception e) {}
			} else {
				if ("Nimbus".equals(laf.getName())) {
					try {
						UIManager.setLookAndFeel(laf.getClassName());
					} catch (Exception e) {}
				}
			}
		}
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				JFrame frame = new DisChoco();
				frame.pack();
				frame.setVisible(true);
			}
		});
	}

	/**
	 * This method initializes generatorsPanel
	 * 
	 * @return javax.swing.JPanel
	 */
	public GeneratorsPanel getGeneratorsPanel() {
		if (generatorsPanel == null) {
			generatorsPanel = new GeneratorsPanel();
		}
		return generatorsPanel;
	}

	/**
	 * This method initializes configurationPanel
	 * 
	 * @return javax.swing.JPanel
	 */
	public ConfigurationPanel getConfigurationPanel() {
		if (configurationPanel == null) {
			configurationPanel = new ConfigurationPanel(new Dimension(1024, 425));
		}
		return configurationPanel;
	}

	/**
	 * This method initializes
	 * 
	 * @return javax.swing.JPanel
	 */
	public ResultsPanel getResultsPanel() {
		if (resultsPanel == null) {
			resultsPanel = new ResultsPanel(getMainFrame(), getPlotButton(), getMergeButton(), getExportAsMenuItem());
		}
		return resultsPanel;
	}

	/**
	 * This method initializes
	 * 
	 * @return javax.swing.JPanel
	 */
	public VisualizingPanel getVisualizingPanel() {
		if (visualizingPanel == null) {
			visualizingPanel = new VisualizingPanel();
		}
		return visualizingPanel;
	}

	/**
	 * This method initializes
	 * 
	 * @return javax.swing.JPanel
	 */
	public SimulationsPanel getSimulationPanel() {
		if (simulationPanel == null) {
			simulationPanel = new SimulationsPanel();
		}
		return simulationPanel;
	}

	/**
	 * This method initializes
	 * 
	 * @return javax.swing.JPanel
	 */
	public DebugPanel getDebugPanel() {
		if (debugPanel == null) {
			debugPanel = new DebugPanel();
		}
		return debugPanel;
	}

	/**
	 * This method initializes myMenuBar
	 * 
	 * @return javax.swing.JMenuBar
	 */
	private JMenuBar getMyMenuBar() {
		if (myMenuBar == null) {
			myMenuBar = new JMenuBar();
			myMenuBar.add(getFileMenu());
			myMenuBar.add(getSimulationMenu());
			myMenuBar.add(getDebugMenu());
			myMenuBar.add(getGeneratorsMenu());
			myMenuBar.add(getExperimentsMenu());
			myMenuBar.add(getResultsMenu());
			myMenuBar.add(getThemesMenu());
			myMenuBar.add(getHelpMenu());
		}
		return myMenuBar;
	}

	/**
	 * This method initializes jMenu
	 * 
	 * @return javax.swing.JMenu
	 */
	private JMenu getFileMenu() {
		if (fileMenu == null) {
			fileMenu = new JMenu();
			fileMenu.setText("File");
			fileMenu.add(getLoadProblemMenuItem());
			fileMenu.add(getLoadResultsMenuItem());
			fileMenu.addSeparator();
			fileMenu.add(getExportAsMenuItem());
			fileMenu.addSeparator();
			fileMenu.add(getExitMenuItem());
		}
		return fileMenu;
	}

	/**
	 * This method initializes jMenu
	 * 
	 * @return javax.swing.JMenu
	 */
	private JMenu getGeneratorsMenu() {
		if (generatorsMenu == null) {
			generatorsMenu = new JMenu();
			generatorsMenu.setText("Generators");
			generatorsMenu.add(getMeetingSchedulingMenuItem());
			generatorsMenu.add(getSensorNetworksMenuItem());
			generatorsMenu.add(getGraphColoringMenuItem());
			generatorsMenu.add(getRandomDisCSPMenuItem());
			generatorsMenu.add(getNQueenMenuItem());
			generatorsMenu.add(getRandomMSPMenuItem());
			generatorsMenu.addSeparator();
			generatorsMenu.add(getComplexRandomDisCSPMenuItem());
			generatorsMenu.add(getComplexGraphColoringMenuItem());
			generatorsMenu.addSeparator();
			generatorsMenu.add(getRandomDCOPMenuItem());
		}
		return generatorsMenu;
	}

	/**
	 * This method initializes experimentsMenu
	 * 
	 * @return javax.swing.JMenu
	 */
	private JMenu getExperimentsMenu() {
		if (experimentsMenu == null) {
			experimentsMenu = new JMenu();
			experimentsMenu.setText("Experiments");
			experimentsMenu.add(getGenerateProblemsMenuItem());
			experimentsMenu.add(getConfigurationMenuItem());
			experimentsMenu.add(getLoadConfigurationFileMenuItem());
		}
		return experimentsMenu;
	}

	/**
	 * This method initializes resultsMenu
	 * 
	 * @return javax.swing.JMenu
	 */
	private JMenu getResultsMenu() {
		if (resultsMenu == null) {
			resultsMenu = new JMenu();
			resultsMenu.setText("Results");
			resultsMenu.add(getPlotResultsMenuItem());
			resultsMenu.add(getMergeResultsMenuItem());
		}
		return resultsMenu;
	}

	/**
	 * This method initializes jMenu
	 * 
	 * @return javax.swing.JMenu
	 */
	private JMenu getHelpMenu() {
		if (helpMenu == null) {
			helpMenu = new JMenu();
			helpMenu.setText("Help");
			helpMenu.add(getAboutMenuItem());
		}
		return helpMenu;
	}

	/**
	 * This method initializes jMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getExitMenuItem() {
		if (exitMenuItem == null) {
			exitMenuItem = new JMenuItem();
			exitMenuItem.setText("Exit");
			exitMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					System.exit(0);
				}
			});
		}
		return exitMenuItem;
	}

	/**
	 * This method initializes jMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getAboutMenuItem() {
		if (aboutMenuItem == null) {
			aboutMenuItem = new JMenuItem();
			aboutMenuItem.setText("About");
			aboutMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					JDialog aboutDialog = getAboutDialog();
					aboutDialog.pack();
					Point loc = getMainFrame().getLocation();
					Point location = getLocation();
					Dimension size = getSize();
					loc.setLocation(size.getWidth() / 2 + location.getX(), size.getHeight() / 2 + location.getY());
					loc.translate(-aboutDialog.getWidth() / 2, -aboutDialog.getHeight() / 2);
					aboutDialog.setLocation(loc);
					aboutDialog.setVisible(true);
				}
			});
		}
		return aboutMenuItem;
	}

	/**
	 * This method initializes aboutDialog
	 * 
	 * @return javax.swing.JDialog
	 */
	private JDialog getAboutDialog() {
		if (aboutDialog == null) {
			aboutDialog = new JDialog(getMainFrame(), true);
			aboutDialog.setTitle("About DisChoco 2.0");
			aboutDialog.setContentPane(getAboutContentPane());
			aboutDialog.setResizable(false);
		}
		return aboutDialog;
	}

	/**
	 * This method initializes aboutContentPane
	 * 
	 * @return javax.swing.JPanel
	 */
	private JPanel getAboutContentPane() {
		if (aboutContentPane == null) {
			aboutContentPane = new JPanel();
			aboutContentPane.setLayout(null);
			aboutContentPane.setPreferredSize(new Dimension(400, 240));
			aboutVersionLabel = new JLabel("DisChoco 2.0", JLabel.CENTER);
			aboutVersionLabel.setBounds(new Rectangle(140, 20, 120, 22));
			logoIconLabel = new JLabel(logoDischoco);
			logoIconLabel.setBounds(new Rectangle(175, 50, 50, 50));
			authorsLabel = new JLabel(authors, JLabel.CENTER);
			authorsLabel.setBounds(new Rectangle(50, 100, 300, 50));
			String webSite = "<html>" +
			"<body>" +
			"<center>" +
			"Web Site:<br>" +
			"<a href=\"http://www.lirmm.fr/coconut/\">http://www.lirmm.fr/coconut/</a>" +
			"</center>" +
			"</body>" +
			"</html>";
			webSiteLabel = new JLabel(webSite, JLabel.CENTER);
			webSiteLabel.setBounds(new Rectangle(50, 150, 300, 50));
			aboutContentPane.add(aboutVersionLabel, null);
			aboutContentPane.add(logoIconLabel, null);
			aboutContentPane.add(authorsLabel, null);  // Generated
			aboutContentPane.add(webSiteLabel, null);  // Generated
		}
		return aboutContentPane;
	}


	/**
	 * This method initializes jMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getLoadProblemMenuItem() {
		if (loadProblemMenuItem == null) {
			loadProblemMenuItem = new JMenuItem();
			loadProblemMenuItem.setText("Load Problem");
			loadProblemMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(java.awt.event.ActionEvent e) {
					String description = "files (xml, bz2)";
					String[] extensions = new String[] {"xml", "bz2"};
					Filter filter = new Filter(extensions, description);
					fileChooser = new JFileChooser(filePath.getPath());
					fileChooser.setFileFilter(filter);
//					fileChooser.setFileView(new JavaFileView());
					fileChooser.setAccessory(new LabelAccessory(fileChooser));
					fileChooser.setPreferredSize(new Dimension(660, 440));
					int returnVal = fileChooser.showOpenDialog(getMainFrame());
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						CardLayout cards = (CardLayout) getNorthPanel().getLayout();
						cards.show(getNorthPanel(), VISUALIZING_PANEL);
						filePath = fileChooser.getSelectedFile();
						println("The file \"" + filePath + "\" was successfelly loaded!");
						visualizingPanel.setProblem();
					} else {
						printlnWarning("No file loaded!");
					}
				}
			});
		}
		return loadProblemMenuItem;
	}

	/**
	 * This method initializes loadResultsMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenuItem getLoadResultsMenuItem() {
		if (loadResultsMenuItem == null) {
			loadResultsMenuItem = new JMenuItem();
			loadResultsMenuItem.setText("Load Results");
			loadResultsMenuItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					CardLayout cards = (CardLayout) getNorthPanel().getLayout();
					cards.show(getNorthPanel(), RESULTS_PANEL);
				}
			});
		}
		return loadResultsMenuItem;
	}

	/**
	 * This method initializes loadResultsMenuItem
	 * 
	 * @return javax.swing.JMenuItem
	 */
	private JMenu getExportAsMenuItem() {
		if (exportAsMenuItem == null) {
			exportAsMenuItem = new JMenu("Export as ..");
			exportAsMenuItem.setEnabled(false);
		}
		return exportAsMenuItem;
	}

	@SuppressWarnings( {"serial", "deprecation"})
	private class JStatusBar extends JPanel implements ActionListener {
		JLabel clock = null;

		public JStatusBar(int width) {
			setLayout(new BorderLayout());
			setPreferredSize(new Dimension(width, 20));
			Timer t = new Timer(1000, this);
			t.start();
			add(getClock(), BorderLayout.WEST);
		}

		public void actionPerformed(ActionEvent ae) {
			getClock().setText((new Date()).toLocaleString());
		}

		private JLabel getClock() {
			if (clock == null) {
				clock = new JLabel("" + new Date().toLocaleString());
			}
			return clock;
		}
	}
	/**
	 * 
	 * @param str
	 */
	public void println(String str) {
		getConsole().println(str);
	}
	/**
	 * 
	 * @param str
	 */
	public void printlnWarning(String str) {
		getConsole().printlnWarning(str);
	}
}
