package kernel.order;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;

import kernel.com.Message;
import util.Commands;
import util.IntArray;
import util.Tools;
import util.graph.Graph;
import util.graph.PriorityTree;

public class TreeAO extends CentralizedAO implements TreeOrder {

	public static int DFSCHAIN = 1;
	public static int BFSCHAIN = 2;

	public static short MinDomDeg 	= 0;
	public static short MaxDeg 		= 1;
	public static short MaxDynDeg 	= 2;
	public static short MinID 		= 3;
	public static short Random 		= 4;
	public static short MaxDegMinDom = 5;
	public static short Heuristic 	= MaxDegMinDom;

	/**
	 * Array will keep track of already chosen agents
	 **/
	protected IntArray visited = null;
	protected IntArray agentsId = null; 
	public int orderSwitch = DFSCHAIN;
	protected PriorityTree pOrderTree = null;
	protected IntArray order = new IntArray();
	protected Comparator<Integer> comparator = null;

	public TreeAO() {
		super();
	}

	/**
	 * 
	 */
	private void init() {
		this.visited = new IntArray();
		this.agentsId = new IntArray();
		for(int agId : this.agentsComplexity.keySet()){
			this.agentsId.add(agId);
		}
		this.comparator = new Comparator<Integer>(){
			@Override
			public int compare(Integer agId1, Integer agId2) {
				int deg1 = agentsComplexity.get(agId1).getNbNeighbors();
				int deg2 = agentsComplexity.get(agId2).getNbNeighbors();
				int dom1 = agentsComplexity.get(agId1).getMaxDomain();
				int dom2 = agentsComplexity.get(agId2).getMaxDomain();
				if( Heuristic == MaxDegMinDom){
					int score = deg2 - deg1;
					if(score == 0){
						score = dom1 - dom2;
					}
					return score;
				} else if(Heuristic == MinDomDeg){
					double score = ((double)dom1)/deg1 - ((double)dom2)/deg2;
					if(score < 0){
						return -1;
					} else if(score > 0) {
						return 1;
					} 
					return 0;
				} else if(Heuristic == MaxDeg){
					return deg2 - deg1;
				} else if(Heuristic == MaxDynDeg){
					return deg2 - deg1;
				} else {
					return 0;
				}
			}
		};
		this.agentsId.sort(this.comparator);
	}
	
	/**
	 * 
	 */
	@Override
	public IntArray buildAgentsOrder() {
		this.init();
		if(Heuristic == MaxDegMinDom){
			this.buildMaxDegMinDomTree();
		} else if(Heuristic == MinDomDeg){
			this.buildMinDomDegTree();
		} else if(Heuristic == MaxDeg){
			this.buildMaxDegTree();
		} else if(Heuristic == MaxDynDeg){
			this.buildMaxDynDegTree();
		} else {
			this.buildRandomTree();
		}
//		System.out.println(this.order);
//		System.out.println(this.pOrderTree);
//		System.out.println(this.pOrderTree.toDFSString());
//		System.exit(0);
		return this.order;
	}
	/**
	 * 
	 */
	//TODO 
	public void buildMaxDegMinDomTree() {
		int root = findNextPriority(-1);
		this.pOrderTree = new PriorityTree(root, this.agentsId.size());
		this.visited.add(root);
		PriorityTree currRoot = pOrderTree;
		while (this.visited.size() < this.agentsId.size()) {/* while we haven't ordered all agents */
			int q = pOrderTree.getCost(currRoot.getRootId()) - 1;
			int agentId = findNextPriority(currRoot.getRootId());
			while (agentId != -1) {/* order as many agents as we can until we hit a leaf */
				PriorityTree s = new PriorityTree(agentId, q);
				this.visited.add(agentId);
				currRoot.addSubTree(s);
				currRoot = s;
				q--;
				agentId = findNextPriority(currRoot.getRootId());
			}
			/* back up one node before restarting loop */
			currRoot = pOrderTree.getParent(currRoot.getRootId());
			if (currRoot == null && pOrderTree.size() != this.agentsId.size()) {
				root = findNextPriority(-1);
				PriorityTree st = new PriorityTree(root, this.agentsId.size() - 2);
				this.visited.add(root);
				pOrderTree.addSubTree(st);
				currRoot = st;
			}
		} /* error check */
		// this.errorCheck();
		this.saveOrder();
	}
	/**
	 * 
	 */
	private void buildRandomTree() {
		Graph<Integer> g = new Graph<Integer>();
		for (int v : this.agentsComplexity.keySet()) {
			g.addVertex(new Integer(v));
			AgentComplexity neig = this.agentsComplexity.get(v);
			for (int w : neig.getNeighbors()) {
				g.addEdge(v, w);
			}
		}
		int root = findNextPriority(-1);
		this.pOrderTree = g.buildDFS(root);
		this.saveOrder();
	}

	/**
	 * 
	 */
	private void buildMaxDynDegTree() {
		Graph<Integer> g = new Graph<Integer>();
		for (int v : this.agentsComplexity.keySet()) {
			g.addVertex(new Integer(v));
			AgentComplexity neig = this.agentsComplexity.get(v);
			for (int w : neig.getNeighbors()) {
				g.addEdge(v, w);
			}
		}
		int root = findNextPriority(-1);
		this.pOrderTree = g.buildMaxDynDegDFS(root);
		this.saveOrder();
	}
	
	public void buildMaxDegTree() {
		try {
			int root = findNextPriority(-1);
			this.pOrderTree = new PriorityTree(root, this.agentsId.size());
			this.visited.add(root);
			PriorityTree currRoot = pOrderTree;
			while (pOrderTree.size() < this.agentsId.size()) {/* while we haven't ordered all agents */
				int q = pOrderTree.getCost(currRoot.getRootId()) - 1;
				int agentId = findNextPriority(currRoot.getRootId());
				while (agentId != -1) {/* order as many agents as we can until we hit a leaf */
					PriorityTree s = new PriorityTree(agentId, q);
					this.visited.add(agentId);
					currRoot.addSubTree(s);
					currRoot = s;
					q--;
					agentId = findNextPriority(currRoot.getRootId());
				}
				/* back up one node before restarting loop */
				currRoot = pOrderTree.getParent(currRoot.getRootId());
				if (currRoot == null && pOrderTree.size() != this.agentsId.size()) {
					root = findNextPriority(-1);
					PriorityTree st = new PriorityTree(root, this.agentsId.size() - 2);
					this.visited.add(root);
					pOrderTree.addSubTree(st);
					currRoot = st;
				}
			} /* error check */
			// this.errorCheck();
			this.saveOrder();
		} catch (Exception e) { e.printStackTrace(); }
	}

	public IntArray getChildren(int id) {
		IntArray childrenId = new IntArray();
		PriorityTree subtree = this.pOrderTree.getSubTree(id);
		for (int i = 0 ; i < subtree.getChildren().size() ; i++) {
			childrenId.add(subtree.getChildren().get(i).getRootId());
		}
		return childrenId;
	}

	// ---------------------------------------------------------------------------------------------------------------------------//

	public int getParentOf(int id) {
		if (this.pOrderTree.getRootId() == id)
			return -1;
		return this.pOrderTree.getParent(id).getRootId();
	}


	// ---------------------------------------------------------------------------------------------------------------------------//

	private void saveOrder() {
		if (this.orderSwitch == BFSCHAIN)
			this.order = IntArray.fromString(pOrderTree.toBFSString());
		else if (this.orderSwitch == DFSCHAIN)
			this.order = IntArray.fromString(pOrderTree.toDFSString());
	}

	// ---------------------------------------------------------------------------------------------------------------------------//

	/**
	 * find an agent Aj such that chosen[Aj] == -1 (it is unchosen), and it has the most links with already chosen agents. 
	 * In case of tie, it should have the most links with unchosen agents. The chosen variable must be connected to the 
	 * given agent with ID = agentId.
	 * @param agId
	 */
	public int findNextPriority(int root) {
		int selected = -1;
		IntArray unvisited = this.agentsId.minus(visited);
		for (int agId : unvisited) {
			/* is it connected to agId1? */
			if (root == -1 || isConnected(root, agId)) {
				if (selected == -1) {
					selected = agId;
				} else {
					if(this.comparator.compare(agId, selected) < 0){
						selected = agId;
					}
				}
			}
		}
		return selected;
	}

	// ---------------------------------------------------------------------------------------------------------------------------//

//	private int compare(int agId1, int agId2) {
//		int deg1 = this.agentsComplexity.get(agId1).getNeighbors().size();
//		int deg2 = this.agentsComplexity.get(agId2).getNeighbors().size();
//		int dom1 = this.agentsComplexity.get(agId1).getMaxDomain();
//		int dom2 = this.agentsComplexity.get(agId2).getMaxDomain();
//		if(Heuristic == MinDomDeg){
//			int domdeg = dom1/deg1 - dom2/deg2;
//			if(domdeg < 0){
//				return -1;
//			} else if(domdeg > 0) {
//				return 1;
//			} else {
//				return 0;
//			}
//		} else if (Heuristic == MaxDeg) {
//			return deg1 - deg2;
//		} else if (Heuristic == MinID) {
//			return agId1 - agId2;
//		} else if (Heuristic == Random) {
//			java.util.Random rnd = new java.util.Random();
//			return rnd.nextInt(3)-1;
//		} else {
//			return 0; //TODO
//		}
//	}

	public boolean isConnected(int agId1, int agId2) {
		IntArray neig1 = this.agentsComplexity.get(agId1).getNeighbors();
		IntArray neig2 = this.agentsComplexity.get(agId2).getNeighbors();
		return neig1.contains(agId2) || neig2.contains(agId1);
	}

	// ---------------------------------------------------------------------------------------------------------------------------//

	public int orderDepth() {
		return pOrderTree.depth();
	}

	// ---------------------------------------------------------------------------------------------------------------------------//

	public PriorityTree getPriorityTree() {
		if (orderSwitch < 4)
			return this.pOrderTree;
		else
			throw new Error(" getPriorityTree() ");

	}

	public HashMap<Integer, IntArray> getNeighborsIds() {
		HashMap<Integer, IntArray> neighborsIds = new HashMap<Integer, IntArray>();
		for (int agId : this.agentsComplexity.keySet()) {
			neighborsIds.put(agId, this.agentsComplexity.get(agId).getNeighbors());
		}
		return neighborsIds;
	}

	/**
	 * 
	 * @param order
	 */
	@Override
	public void broadcastOrder() {
		PriorityTree ptree = this.getPriorityTree();
		for (int receiverId : this.getAgentsComplexity().keySet()) {
			int parentId = this.getParentOf(receiverId);
			IntArray childrenIds = this.getChildren(receiverId);
			HashMap<Integer, IntArray> neighbs = this.getNeighborsIds();
			Message msg = createMessage(Commands.AO);
			msg.addContents(Commands.ORDER, order);
			msg.addContents(Commands.GRAPH, Graph.toMsg(graph));
			msg.addContents("Parent-Id", parentId + "");
			msg.addContents("Children-Ids", childrenIds);
			msg.addContents("Priority-tree", ptree.toString());
			msg.addContents("Neighbors", neighbs.toString());
			msg.setReceiverId(receiverId);
			this.master.sendMessage(msg);
		}
	}

	public String pretty() {
		String treeAO = "";
		PriorityTree ptree = this.getPriorityTree();
		treeAO += ptree.pretty();
		return treeAO;
	}

	public String toString() {
		String treeAO = "";
		PriorityTree ptree = this.getPriorityTree();
		treeAO += ptree.toString();
		return treeAO;
	}
	/**
	 * 
	 * @param complexities
	 * @return
	 */
	public static PriorityTree buildpriorityTree(HashMap<Integer, AgentComplexity> complexities) {
		int nbAgents = complexities.size();
		IntArray visited = new IntArray();
		HashMap<Integer, Double> metric = computeMetric(complexities);
		int root = -1;
		if(Heuristic == MaxDeg || Heuristic == MaxDynDeg ){
			root = Tools.MaxOf(metric);
		} else {
			root = Tools.MinOf(metric); //findNextPriority(-1, complexities, visited, metric);
		}
		metric.remove(root);
		PriorityTree pOrderTree = new PriorityTree(root, nbAgents);
		visited.add(root);
		PriorityTree currRoot = pOrderTree;
		/* while we haven't ordered all agents */
		while (pOrderTree.size() < nbAgents) {
			int q = pOrderTree.getCost(currRoot.getRootId()) - 1;
			int agentId = findNextPriority(currRoot.getRootId(), complexities, visited, metric);
			while (agentId != -1) {/* order as many agents as we can until we hit a leaf */
				PriorityTree s = new PriorityTree(agentId, q);
				visited.add(agentId);
				metric.remove(agentId);
				currRoot.addSubTree(s);
				currRoot = s;
				q--;
				agentId = findNextPriority(currRoot.getRootId(), complexities, visited, metric);
			}
			/* back up one node before restarting loop */
			currRoot = pOrderTree.getParent(currRoot.getRootId());
			if (currRoot == null && pOrderTree.size() != nbAgents) {
				root = findNextPriority(-1, complexities, visited, metric);
				PriorityTree st = new PriorityTree(root, nbAgents - 2);
				visited.add(root);
				metric.remove(root);
				pOrderTree.addSubTree(st);
				currRoot = st;
			}
		}
		return pOrderTree; 
//		//TODO
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//		Graph<Integer> g = new Graph<Integer>();
//		for (int v : complexities.keySet()) {
//			g.addVertex(new Integer(v));
//			AgentComplexity neig = complexities.get(v);
//			for (int w : neig.neighbors) {
//				g.addEdge(v, w);
//			}
//		}
//		// root = findNextPriority(-1, complexities, visited);
//		PriorityTree tree = null;
//		if(Heuristic == MaxDeg){
//			tree = g.buildDFS(root);
//		} else if(Heuristic == MaxDynDeg){
//			tree = g.buildMaxDynDegDFS(root);
//		} else {
//			tree = g.buildDFS(root);
//		}
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//		return tree;
	}

	private static int findNextPriority(int agId1, HashMap<Integer, AgentComplexity> complexities, IntArray visited, HashMap<Integer, Double> metric) {
		int nextId = -1;
//		Iterator<Integer> it = complexities.keySet().iterator();
//		while (it.hasNext()) {
//			int agId2 = it.next();
		for (int agId2 : metric.keySet()){
			/* is this agent currently unchosen? */
			if (!visited.contains(agId2)) {
				/* is it connected to agId1? */
				if (agId1 == -1 || isConnected(agId1, agId2, complexities)) {
					if (nextId == -1) {
						nextId = agId2;
					} else {
						// this.visited.add(agId2);
						if (compare(agId2, nextId, metric) > 0)
							nextId = agId2;
					}
				}
			}
		}
		return nextId;
	}

	/**
	 * @param complexities
	 * @return
	 */
	private static HashMap<Integer, Double> computeMetric(HashMap<Integer, AgentComplexity> complexities) {
		HashMap<Integer, Double> operator = new HashMap<Integer, Double>();
		Iterator<Integer> it = complexities.keySet().iterator();
		while (it.hasNext()) {
			int agId = it.next();
			AgentComplexity agComp = complexities.get(agId);
			double dom = agComp.getMaxDomain();
			double deg = agComp.getNeighbors().size();
			if(Heuristic == MinDomDeg) {
				if (deg == 0){
					operator.put(agId, Double.POSITIVE_INFINITY);
				} else{
					operator.put(agId, dom / deg);
				}
			} else if(Heuristic == MaxDeg){
				operator.put(agId, deg);
			} else if(Heuristic == MaxDynDeg){
				operator.put(agId, deg);
			} else if(Heuristic == MinID){
				operator.put(agId, (double) agId);
			} else {
				operator.put(agId, Math.random());
			}
		}
		return operator;
	}

	private static int compare(int agId1, int agId2, HashMap<Integer, Double> metric) {
		double op = metric.get(agId1) - metric.get(agId2);
		if (op < 0){
			return -1 ;
		} else if(op > 0){
			return 1;
		}
		return 0;
	}

	private static boolean isConnected(int agId1, int agId2, HashMap<Integer, AgentComplexity> complexities) {
		IntArray neig1 = complexities.get(agId1).getNeighbors();
		IntArray neig2 = complexities.get(agId2).getNeighbors();
		return Tools.findInArray(neig1.getArray(), agId2) != -1 || Tools.findInArray(neig2.getArray(), agId1) != -1;
	}
	
	/**
	 * 
	 */
	//TODO 
	public void buildMinDomDegTree() {
		try {
			int root = findNextPriority(-1);
			this.pOrderTree = new PriorityTree(root, this.agentsId.size());
			this.visited.add(root);
			PriorityTree currRoot = pOrderTree;
			while (pOrderTree.size() < this.agentsId.size()) {/* while we haven't ordered all agents */
				int q = pOrderTree.getCost(currRoot.getRootId()) - 1;
				int agentId = findNextPriority(currRoot.getRootId());
				while (agentId != -1) {/* order as many agents as we can until we hit a leaf */
					PriorityTree s = new PriorityTree(agentId, q);
					this.visited.add(agentId);
					currRoot.addSubTree(s);
					currRoot = s;
					q--;
					agentId = findNextPriority(currRoot.getRootId());
				}
				/* back up one node before restarting loop */
				currRoot = pOrderTree.getParent(currRoot.getRootId());
				if (currRoot == null && pOrderTree.size() != this.agentsId.size()) {
					root = findNextPriority(-1);
					PriorityTree st = new PriorityTree(root, this.agentsId.size() - 2);
					this.visited.add(root);
					pOrderTree.addSubTree(st);
					currRoot = st;
				}
			} /* error check */
			// this.errorCheck();
			this.saveOrder();
		} catch (Exception e) { e.printStackTrace(); }
	}

	public int compare(Integer agent1, Integer agent2) {
		if( Heuristic == MaxDegMinDom){
			int deg1 = this.agentsComplexity.get(agent1).getNbNeighbors();
			int deg2 = this.agentsComplexity.get(agent2).getNbNeighbors();
			int score = deg1 - deg2;
			if(score == 0){
				int dom1 = this.agentsComplexity.get(agent1).getMaxDomain();
				int dom2 = this.agentsComplexity.get(agent2).getMaxDomain();
				return dom1 - dom2;
			}
			return score;
		}
		return 0;
	}
	
	/**
	 * 
	 */
//	private void buildComplexTree() {
//		this.nbAgents = this.agentsComplexity.keySet().size();
//		int root = selectRoot();
//		this.pOrderTree = new PriorityTree(root, this.nbAgents);
//		this.visited.add(root);
//		PriorityTree currRoot = pOrderTree;
//		
////		IntArray localAgents = this.agentsComplexity.get(root).getLocalAgents();
////		localAgents = localAgents.minus(this.visited);
////		while(!this.visited.contains(localAgents)){
////			HashMap<Integer, Double> local_dom_deg = new HashMap<Integer, Double>();
////			for(int id : localAgents){
////				local_dom_deg.put(id, dom_deg.get(id));
////			}
////			int localId = Tools.MinOf(local_dom_deg);
////			torder.add(localId);
////			dom_deg.remove(localId);
////			localAgents.remove(localId);
////		}
//		
//		while (pOrderTree.size() < nbAgents) {/* while we haven't ordered all agents */
//			int q = pOrderTree.getCost(currRoot.getRootId()) - 1;
//			int agentId = findNextPriority(currRoot.getRootId());
//			while (agentId != -1) {/* order as many agents as we can until we hit a leaf */
//				PriorityTree s = new PriorityTree(agentId, q);
//				this.visited.add(agentId);
//				currRoot.addSubTree(s);
//				currRoot = s;
//				q--;
//				agentId = findNextPriority(currRoot.getRootId());
//			}
//			/* back up one node before restarting loop */
//			currRoot = pOrderTree.getParent(currRoot.getRootId());
//			if (currRoot == null && pOrderTree.size() != nbAgents) {
//				root = findNextPriority(-1);
//				PriorityTree st = new PriorityTree(root, nbAgents - 2);
//				this.visited.add(root);
//				pOrderTree.addSubTree(st);
//				currRoot = st;
//			}
//		} /* error check */
//		this.saveOrder();
//		System.out.println(this.order);
//		System.exit(0);
//	}


//	/**
//	 * @return
//	 */
//	private Graph<Integer> buildGraph() {
//		Graph<Integer> g = new Graph<Integer>();
//		for(Integer agId : this.agentsComplexity.keySet()){
//			g.addVertex(agId);
//			for(Integer neig : this.agentsComplexity.get(agId).getNeighbors()){
//				g.addEdge(agId, neig);
//			}
//		}
//		return g;
//	}
}
