package simple.dcsp.afcfamily.afc.estimDom;

import java.util.HashMap;

import kernel.StopException;
import kernel.com.Message;
import kernel.events.VariableEvent;
import kernel.order.AgentOrdering;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.dcsp.afcfamily.afc.AFCAgent;
import util.Commands;
import util.Instantiation;
import util.IntArray;
import util.VarEventMask;

public class AFCEstimatedAgent extends AFCAgent {

	private static final short MinDom = 1;
	private static final short MinDomDeg = 2;

	private short type;

	long myCounter = 0;
	HashMap<Integer, Integer> lbDomais;
	HashMap<Integer, Integer> ubDomais;

	HashMap<Integer, Integer> LB;
	HashMap<Integer, Integer> UB;

	public AFCEstimatedAgent(int agentId, String agName, String description, short type) {
		super(agentId, agName, description);
		this.type = type;
		this.lbDomais = new HashMap<Integer, Integer>();
		this.ubDomais = new HashMap<Integer, Integer>();
		this.cpaMsgH = new EstCPAMsgHandler(this);
		this.agentOrder = new AgentOrdering(this) {
			public void processOrder(Message msg) {
				super.processOrder(msg);
				for (int id : this.agentsOrder) {
					lbDomais.put(id, (Integer) msg.get(id + ""));
					ubDomais.put(id, (Integer) msg.get(id + ""));
				}
			}
		};
	}

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_FIRST).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_FIRST).add(VariableEvent.DESINSTANCIATE);
		super.configurePropagationEngine();
	}

	// ==========================================================================

	public void sendCPA(Instantiation CPA) throws StopException {
		this.computeBounds(CPA);
		super.sendCPA(CPA);
	}

	// ==========================================================================
	protected void receiveCPA(Instantiation cpa, long sc, boolean fromBackMsg) throws StopException {
		if (sc > this.stepCounter) {
			if (this.updateAgentOrdering(cpa, !fromBackMsg)) {
				this.agentView.clearExcept(cpa);
				// this.agentView.priorityMerge(notOk);
				this.agentViewConsistent = true;
			}
		}
		super.receiveCPA(cpa, sc, fromBackMsg);
	}

	// ==========================================================================
	public void forwardCheck(Instantiation cpa, long sc) {
		if (sc > this.stepCounter) {
			if (this.updateAgentOrdering(cpa, false)) {
				this.agentView.clearExcept(cpa);
				// this.agentView.priorityMerge(cpa);
				this.agentViewConsistent = true;
			}
			super.forwardCheck(cpa, sc);
		}
	}

	// ==========================================================================
	public void processNotOk(Instantiation notOk, long sc) {
		if (sc > this.stepCounter) {
			if (this.updateAgentOrdering(notOk, false)) {
				this.agentView.clearExcept(notOk);
			}
		}
		super.processNotOk(notOk, sc);
	}

	// ==========================================================================
	public boolean updateAgentOrdering(Instantiation cpa, boolean ImNext) {
		IntArray newOrder = cpa.getAgentIds();
		if (ImNext)
			newOrder.add(this.agentId);
		IntArray others = this.agentOrder.getOrder().minus(newOrder);
		newOrder.addAll(others);
		if (this.agentOrder.setNewOrder(newOrder)) {
			this.agentView.ensureOnlyHigherAgentsInView();
			this.constraintHandler.update();
			return true;
		}
		return false;

	}

	// ==========================================================================

	protected void computeBounds(Instantiation cpa) {
		this.LB = new HashMap<Integer, Integer>();
		this.UB = new HashMap<Integer, Integer>();
		IntArray unassigned = this.agentOrder.getOrder();
		unassigned = unassigned.minus(cpa.getAgentIds());
		for (int id : unassigned) {
			IntExternalVar exV = (IntExternalVar) this.getExternalVar(id);
			if (exV != null && this.communicator.isNeighbor(id)) {
				int ub = this.ubDomais.get(id);
				UB.put(id, Math.min(ub, exV.getDomainSize()));
				int conflict = exV.getOriginalDomainSize() - exV.getDomainSize();
				int lb = this.lbDomais.get(id) - conflict;
				LB.put(id, Math.min(exV.getDomainSize(), lb));

			} else {
				LB.put(id, this.lbDomais.get(id));
				UB.put(id, this.ubDomais.get(id));
			}
		}
	}

	// ====================================================================================
	public int getNextAgent() {
		switch (this.type) {
			case MinDom :
				return this.getMinDomainNextAgent();
			case MinDomDeg :
				return this.getDomDegNextAgent();
		}
		return -1;
	}

	/**
	 * @param bounds
	 * @return next agent to assign
	 */
	public int getMinDomainNextAgent() {
		if (UB.isEmpty())
			return -1;
		for (int agId : UB.keySet()) {
			if (UB.get(agId) == 0) {
				return agId;
			} else {
				boolean selected = true;
				for (int id : LB.keySet()) {
					if (agId != id) {
						if (UB.get(agId) > LB.get(id)) {
							selected = false;
							break;
						}
					}
				}
				if (selected)
					return agId;
			}
		}
		int IdOfMinLb = -1;
		int minLb = Integer.MAX_VALUE;
		for (int agId : LB.keySet()) {
			if (LB.get(agId) < minLb) {
				IdOfMinLb = agId;
				minLb = LB.get(agId);
			}
		}
		return IdOfMinLb;
	}

	// ===============================================================================
	/**
	 * @param bounds
	 * @return next agent to assign
	 */
	public int getDomDegNextAgent() {
		IntArray assigned = this.agentView.toInstantiation().getAgentIds();
		assigned.add(this.agentId);
		IntArray unassigned = this.agentOrder.getOrder().minus(assigned);
		if (unassigned.size() == 0 || UB.isEmpty())
			return -1;
		HashMap<Integer, Integer> degrees = this.getDegrees(unassigned);

		for (int agId : UB.keySet()) {
			int deg = degrees.get(agId);
			double upperBound = UB.get(agId);
			if (upperBound == 0) {
				return agId;
			} else {
				if (deg == 0)
					upperBound = Double.POSITIVE_INFINITY;
				else
					upperBound /= deg;
				boolean selected = true;
				for (int id : LB.keySet()) {
					int degree = degrees.get(id);
					double lowerBound = LB.get(id);
					if (degree == 0)
						lowerBound = Double.POSITIVE_INFINITY;
					else
						lowerBound /= degree;
					if (agId != id) {
						if (upperBound > lowerBound) {
							selected = false;
							break;
						}
					}
				}
				if (selected) { return agId; }
			}
		}
		int IdOfMinLb = -1;
		double minLb = Double.POSITIVE_INFINITY;
		for (int agId : LB.keySet()) {
			int deg = degrees.get(agId);
			double lowerBound = LB.get(agId);
			if (deg == 0)
				lowerBound = Double.POSITIVE_INFINITY;
			else
				lowerBound /= deg;
			if (lowerBound < minLb) {
				IdOfMinLb = agId;
				minLb = LB.get(agId);
			}
		}
		return IdOfMinLb;
	}

	// ===============================================================================
	private HashMap<Integer, Integer> getDegrees(IntArray unassigned) {
		HashMap<Integer, Integer> map = new HashMap<Integer, Integer>();
		for (int agId : UB.keySet()) {
			map.put(agId, ddegree(agId, unassigned));
		}
		return map;
	}

	// ===============================================================================
	private int ddegree(int agId, IntArray unassigned) {
		int deg = 0;
		Iterable<Integer> neighbors = this.agentOrder.getAgentsGraph().getNeighbors(agId);
		for (int id : neighbors)
			if (unassigned.contains(id)) {
				deg++;
			}
		return deg;
	}

	// ===============================================================================

	public class EstCPAMsgHandler extends CPAMsgHandler {

		public EstCPAMsgHandler(SimpleAgent algo) {
			super(algo);
		}

		@SuppressWarnings("unchecked")
		public void processMessage(Message msg) throws StopException {
			HashMap<Integer, Integer> lbDomais = (HashMap<Integer, Integer>) msg.get("LB");
			HashMap<Integer, Integer> ubDomais = (HashMap<Integer, Integer>) msg.get("UB");
			lbDomais.putAll(lbDomais);
			ubDomais.putAll(ubDomais);
			super.processMessage(msg);
		}

		public void sendCPAMessage(Instantiation cpa, int receiverId, long sc) {
			Message msg = createMessage();
			msg.addContents(Commands.CPA, cpa.clone());
			msg.addContents("SC", sc);
			msg.addContents("LB", LB);
			msg.addContents("UB", UB);
			msg.setReceiverId(receiverId);
			this.myAgent.sendMessage(msg);
		}
	}
}
