package simple.dcsp.afcfamily.afcng;

import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.vars.integer.IntExternalVar;
import simple.Nogood;
import util.Commands;
import util.Instantiation;
import util.IntArray;
import util.VarAssignment;

public class DynamicAFCngAgent extends AFCngAgent {
	
	/**
	 * 
	 * @param agentId
	 * @param agName
	 * @param description
	 */
	public DynamicAFCngAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.ngdMsgH = new DynNogoodMsgHandler(this);
	}
	
	/**
	 * 
	 */
	public void initAgentView() {
		IntArray agents = this.getOrder();
		agents.remove(this.agentId);
		for (int agId : agents) {
			if (!this.communicator.isNeighbor(agId)) {
				VarAssignment var = new VarAssignment(new VarIdentifier(agId), Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(var);
				this.agentView.add(extVar);
			} else {
				this.agentView.add((IntExternalVar) this.getExternalVar(agId));
			}
		}
	}
	
	/**
	 * 
	 */
	public void receiveCPA(TimeStamp ts, Instantiation cpa, int nextId) throws StopException {
		if (this.timeStamp.compareTimeStamp(ts) > 0) {
			this.timeStamp.setTimeStamp(ts);
			IntArray newOrdering = cpa.getAgentIds();
			newOrdering.add(nextId);
			this.agentOrder.updateOrder(newOrdering);
			this.constraintHandler.update();
			this.processCPA(cpa, nextId);
		} else {
			this.auxiliaryStatistics.get(OBSOLETE_CPA).value++;
		}
	}
	
	/**
	 * 
	 * @param nogood
	 * @param timeStamp
	 * @param cpa
	 * @throws StopException
	 */
	public void processNogood(Nogood nogood, TimeStamp timeStamp, Instantiation cpa, int generator) throws StopException {
		int cmp = this.timeStamp.compareTimeStamp(timeStamp);
		if (cmp > 0) {
			this.updateOrdering(cpa, timeStamp);
		}
		if (cmp >= 0) {
			if(this.agentViewConsistent && this.isCoherent(nogood)){
				this.makeNoGoodConstraint(nogood);
				this.assignCPA();
			} else{
				this.auxiliaryStatistics.get(Commands.nbObsoleteNogood).value++;
			}
		}
	}
	
	/**
	 * @param cpa
	 * @param timeStamp
	 * @throws StopException 
	 */
	public void updateOrdering(Instantiation cpa, TimeStamp timeStamp) throws StopException {
		this.timeStamp.setTimeStamp(timeStamp);
		this.agentViewConsistent = true;
		this.agentOrder.updateOrder(cpa.getAgentIds());
		this.constraintHandler.update();
		if (!this.updateAgentView(cpa)) {
			this.backtrack();
		}
	}

	protected class DynNogoodMsgHandler extends NogoodMsgHandler {
		public DynNogoodMsgHandler(DynamicAFCngAgent agent) {
			super(agent);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Nogood nogood = (Nogood) msg.get(NOGOOD);
			Instantiation cpa = (Instantiation) msg.get(CPA);
			TimeStamp timeStamp = (TimeStamp) msg.get(TIMESTAMP);
			processNogood(nogood, timeStamp, cpa, msg.getSenderId());
		}

		@Override
		public void sendNogood(Nogood nogood) {
			int target = nogood.getRhsAgentId();
			Instantiation cpa = agentView.getInstantiationOfAgentsHigherThan(target);
			Message msg = this.createMessage();
			msg.addContents(NOGOOD, nogood);
			msg.addContents(CPA, cpa);
			msg.addContents(TIMESTAMP, timeStamp.clone());
			msg.setReceiverId(target);
			sendMessage(msg);
		}
	}

}