package simple.dcsp.afcfamily.afcng.afctree;

import kernel.StopException;
import kernel.VarIdentifier;
import kernel.order.TreeDisAO;
import kernel.vars.integer.IntExternalVar;
import simple.dcsp.afcfamily.afcng.AFCngAgent;
import util.Instantiation;
import util.IntArray;
import util.VarAssignment;

public class AFCTreeAgent extends AFCngAgent {

	private int myLevel;

	public AFCTreeAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.terminateMsgHandler = new AFCTreeTermMsgHandler(this);
		this.agentOrder = new TreeDisAO(this);
	}

	// ==========================================================================

	public void initAgentView() {
		IntArray agents = ((TreeDisAO) this.agentOrder).getAncestors();
		for (int agId : agents) {
			VarIdentifier varId = new VarIdentifier(agId);
			VarAssignment var = new VarAssignment(varId, Integer.MIN_VALUE);
			IntExternalVar extVar = this.makeExternalVar(var);
			this.agentView.add(extVar);
		}
		this.myLevel = agents.size();
	}

	// ==========================================================================
	/**
	 * Start distributed search from the initializing agent
	 * 
	 * @throws StopExeption
	 */
	public void startDistributedSearch() throws StopException {
		if (((TreeDisAO) this.agentOrder).getParentId() == -1) {
			this.assignCPA();
		}
	}

	// ==========================================================================

	public void checkAssign(int agentId) throws StopException {
		if (this.isMyParent(agentId)) {
			this.assignCPA();
		}
	}

	// ==========================================================================

	private boolean isMyParent(int agId) {
		return agId == ((TreeDisAO) this.agentOrder).getParentId();
	}

	// ==========================================================================

	public void sendCPA(Instantiation cpa) throws StopException {
		IntArray linkedDescendants = this.communicator.getLowerNeighbors();
//		IntArray descendants = ((TreeDisAO) this.agentOrder).getDescendants();
		this.timeStamp.setTimeStamp(cpa.getTimeStamp());
		for (int receiverId : linkedDescendants) {
			this.cpaMsgH.sendCPAMessage(cpa, receiverId, this.agentId);
		}
		if (linkedDescendants.size() == 0) {
			((AFCTreeTermMsgHandler) this.terminateMsgHandler).sendAccepted();
		}
	}

	public void incrementMyStamp() {
		this.timeStamp.update(myLevel, this.var.getTimeCounter());
	}

}
