/**
 * 
 */
package simple.dcsp.agile.exp;

import static simple.dcsp.agile.exp.AgileMaster.DomDeg;
import static simple.dcsp.agile.exp.AgileMaster.DomWDeg;
import static simple.dcsp.agile.exp.AgileMaster.MinDom;
import static simple.dcsp.agile.exp.AgileMaster.Topological_Deg;
import static simple.dcsp.agile.exp.AgileMaster.Topological_ID;
import static simple.dcsp.agile.exp.AgileMaster.Topological_MaxSucc;

import java.util.BitSet;
import java.util.HashMap;

import util.Instantiation;
import util.IntArray;

public class Explanations {
	private BitSet[] lhs = null;
	private int[] rhs = null;
	private int[] domains = null;
	private int[] degrees = null;
	private int[] weights = null;
	private IntArray agentIds = null;
	private IntArray newOrder = null;
	private double[] newTV = null;
	private Instantiation newNogood = null;
	private short heuristic = MinDom;
	
	public Explanations(IntArray agentIds, int[] domains, int[] degrees, int[] weights, short heuristic) {
		this.heuristic = heuristic;
		this.agentIds = agentIds.clone();
		this.domains = domains.clone();
		this.degrees = degrees.clone();
		this.weights = weights.clone();
		this.rhs = new int[agentIds.size()+1];
		this.lhs = new BitSet[agentIds.size()+1];
		for(int agId : agentIds){
			this.lhs[agId] = new BitSet(agentIds.size()+1);
			this.removeExplanationOf(agId);
		}
	}
	/**
	 * 
	 */
	public Explanations clone() {
		IntArray agentIds = this.agentIds.clone();
		return new Explanations(agentIds, this.domains.clone(), this.degrees.clone(), weights.clone(), 
				this.heuristic, this.getLhsClone(), this.rhs.clone());
	}

	/**
	 * @return
	 */
	private BitSet[] getLhsClone() {
		BitSet[] clone = new BitSet[this.lhs.length];
		for(int agId : agentIds){
			clone[agId] = (BitSet) this.lhs[agId].clone();
		}
		return clone;
	}
	/**
	 * 
	 * @param agentIds
	 * @param maxDomSize
	 * @param degrees
	 * @param heuristic
	 * @param agentView
	 * @param lhs
	 * @param rhs
	 */
	public Explanations(IntArray agentIds, int[] domains, int[] degrees, int[] weights, short heuristic, BitSet[] lhs, int[] rhs) {
		this.heuristic = heuristic;
		this.agentIds = agentIds;
		this.domains = domains;
		this.degrees = degrees;
		this.weights = weights;
		this.lhs = lhs;
		this.rhs = rhs;
	}

	public void updateMaxDomainSize(int agId, int d) {
		int value = Math.min(this.getMaxDomainSize(agId), d);
		this.domains[agId] =  value;
	}

	public int getMaxDomainSize(int agId) {
		return this.domains[agId];
	}

	public void removeExplanationOf(int agId) {
		this.lhs[agId].clear();
		this.rhs[agId] = this.domains[agId];
	}

	public IntArray getLhs(int agId) {
		IntArray predecessor = new IntArray();
		for(int i=this.lhs[agId].nextSetBit(0) ; i>-1 ; i=this.lhs[agId].nextSetBit(i+1)){
			predecessor.add(i);
		}
		return predecessor;
	}

	public int getRhs(int agId) {
		return this.rhs[agId];
	}
	
	public void setRhs(int agId, int rhs) {
		this.rhs[agId] = rhs;
	}
	
	public void setExplanation(int agId, IntArray lhs, int rhs) {
		this.setRhs(agId, rhs);
		this.setLhs(agId, lhs);
	}

	/**
	 * @param agId
	 * @param lhs
	 */
	private void setLhs(int agId, IntArray lhs) {
		this.lhs[agId].clear();
		for(int pre : lhs){
			this.lhs[agId].set(pre);
		}
		if(this.lhs[agId].get(0)){
			throw new Error("\nagId: "+agId+"  <--> lhs: "+lhs+"\n"+ this.toString()+"\n");
		}
	}

	public IntArray getNewOrder() {
		return this.newOrder;
	}

	public void setNewOrder(IntArray newOrder) {
		this.newOrder = newOrder;
	}

	public double[] getNewTV() {
		return this.newTV;
	}

	public void setNewTV(double[] newTV) {
		this.newTV = newTV;
	}

	public Instantiation getNewNG() {
		return this.newNogood;
	}

	public void setNewNG(Instantiation newNG) {
		this.newNogood = newNG;
	}

	public void clear() {
		this.newOrder = null;
		this.newTV = null;
		this.newNogood = null;
	}

	/**
	 * remove all explanations incompatible with order
	 */
	public int removeIncompatibleLhsWithOrder(IntArray order) {
		int nbChecks = 0;
		for(int agId=1 ; agId<this.lhs.length ; agId++){
			IntArray higherAgents = order.getElementsBefor(agId);// this.getHigherAgents(order, agId);
			IntArray predecessors = this.getLhs(agId);
			if (!higherAgents.contains(predecessors)) {
				this.removeExplanationOf(agId);
			}
			nbChecks++;
		}
		return nbChecks;
	}
	
	/**
	 * remove all explanations incompatible with order
	 */
	public void removeExplanationOfAgents(IntArray agents) {
		for (int agId : agents) {
			this.removeExplanationOf(agId);
		}
	}

	public IntArray getTopologicalOrder(short topologicalType) {
		IntArray order = new IntArray();
		HashMap<Integer, IntArray> successors = this.getGraphSuccessor();
		HashMap<Integer, IntArray> graph = this.getGraphPredecessor();
		while (graph.keySet().size() > 0) {
			IntArray roots = this.getRoots(graph);
			int selected = -1;
			double min = Integer.MAX_VALUE;
			double rhs = 0;
			switch (topologicalType) {
				case Topological_ID : {
					for (int agId : roots) {
						if (this.heuristic == MinDom) {
							rhs = this.getRhs(agId);
						} else if (this.heuristic == DomDeg) {
							rhs = ((double) this.rhs[agId]) / this.degrees[agId];
						} else if (this.heuristic == DomWDeg) {
							rhs = ((double) this.rhs[agId]) / (this.weights[agId]*this.degrees[agId]);
						}
						if ((min > rhs) || ((min == rhs) && (agId < selected))) {
							min = this.getRhs(agId);
							selected = agId;
						}
					}
					break;
				}
				case Topological_Deg : {
					int deg = Integer.MIN_VALUE;
					for (int agId : roots) {
						if (this.heuristic == MinDom) {
							rhs = this.rhs[agId];
						} else if (this.heuristic == DomDeg) {
							rhs = ((double) this.rhs[agId]) / this.degrees[agId];
						} else if (this.heuristic == DomWDeg) {
							rhs = ((double) this.rhs[agId]) / (this.weights[agId]*this.degrees[agId]);
						}
						if (min > rhs || (min == rhs && (this.degrees[agId] > deg))) {
							min = this.rhs[agId];
							deg = this.degrees[agId];
							selected = agId;
						}
					}
					break;
				}
				case Topological_MaxSucc : {
					for (int agId : roots) {
						if (this.heuristic == MinDom) {
							rhs = this.getRhs(agId);
						} else if (this.heuristic == DomDeg) {
							rhs = ((double) this.rhs[agId]) / this.degrees[agId];
						} else if (this.heuristic == DomWDeg) {
							rhs = ((double) this.rhs[agId]) / (this.weights[agId]*this.degrees[agId]);
						}
						if (min > rhs || (min == rhs && successors.get(agId).size() > successors.get(selected).size())) {
							min = this.getRhs(agId);
							selected = agId;
						}
					}
					break;
				}
			}
			order.add(selected);
			this.remove(graph, selected);
		}
//		if (order.size() != this.agentIds.size()) { throw new Error("\nacyclic graph : \n" + this); }
		return order;
	}
	/**
	 * 
	 * @param logger 
	 * @return
	 */
	public IntArray getBestOrder() {
		IntArray order = new IntArray();
		HashMap<Integer, IntArray> graph = this.getGraphPredecessor();
		while (order.size() < this.agentIds.size()) {
			IntArray allRoots = this.getRoots(graph);
			IntArray S1 = this.getMinRoots(allRoots);
			while (S1.size() != 0) {
				if (S1.size() > 1) {
					int d = this.getRhs(S1.get(0));
					IntArray S2 = this.getAgentsWithMinRhs(graph, d);
					if (S2.size() == 0) {
						int selected = S1.get(0);
						order.add(selected);
						this.remove(graph, selected);
						break;
					}
					IntArray S3 = this.getRootsWithMinPredecessor(graph, S2);
					IntArray S4 = this.getMinRoots(S3);
					if (S4.size() == 1) {
						int selected = S4.get(0);
						IntArray predecessors = graph.get(selected);
						for (int pred : graph.get(selected)) {
							predecessors.addAll(graph.get(pred));
						}
						while (predecessors.size() > 0) {
							int pred = predecessors.get(0);
							predecessors.remove(pred);
							if (order.contains(graph.get(pred))) {
								order.add(pred);
								this.remove(graph, pred);
							} else {
								predecessors.add(pred);
							}
						}
						order.add(selected);
						this.remove(graph, selected);
						break;
					} else {
						S1 = S4;
					}
				} else {
					int selected = S1.get(0);
					order.add(selected);
					this.remove(graph, selected);
					break;
				}
			}
		}
		return order;
	}

	/**
	 * 
	 * @param graph
	 * @param agents
	 * @return
	 */
	private IntArray getRootsWithMinPredecessor(HashMap<Integer, IntArray> graph, IntArray agents) {
		int min = Integer.MAX_VALUE;
		for (int agId : agents) {
			if (graph.get(agId).size() < min) {
				min = graph.get(agId).size();
			}
		}
		IntArray ans = new IntArray();
		for (int agId : agents) {
			if (graph.get(agId).size() == min) {
				ans.add(agId);
			}
		}
		return ans;
	}

	/**
	 * 
	 * @param graph
	 * @param d
	 * @return
	 */
	private IntArray getAgentsWithMinRhs(HashMap<Integer, IntArray> graph, int d) {
		IntArray agents = new IntArray();
		for (int agId : graph.keySet()) {
			if (this.getRhs(agId) < d) {
				agents.add(agId);
				for (int pre : graph.get(agId)) {
					if (this.getRhs(pre) != d) {
						agents.remove(agId);
						break;
					}
				}
			}
		}
		return agents;
	}

	/**
	 * 
	 * @param roots
	 * @return
	 */
	private IntArray getMinRoots(IntArray allRoots) {
		int min = Integer.MAX_VALUE;
		for (int agId : allRoots) {
			if (this.getRhs(agId) < min) {
				min = this.getRhs(agId);
			}
		}
		IntArray roots = new IntArray();
		for (int agId : allRoots) {
			if (this.getRhs(agId) == min) {
				roots.add(agId);
			}
		}
		return roots;
	}

	/**
	 * 
	 * @param graph
	 * @param selected
	 */
	private void remove(HashMap<Integer, IntArray> graph, int selected) {
		for (int agId : graph.keySet()) {
			if (graph.get(agId).contains(selected)) {
				graph.get(agId).remove(selected);
			}
		}
		graph.remove(selected);
	}

	/**
	 * 
	 * @param graph
	 *            is the graph of predecessors
	 * @return
	 */
	private IntArray getRoots(HashMap<Integer, IntArray> graph) {
		IntArray roots = new IntArray();
		for (int agId : graph.keySet()) {
			if ((graph.get(agId).size() == 0)) {
				roots.add(agId);
			}
		}
		return roots;
	}

	/**
	 * 
	 * @return
	 */
	private HashMap<Integer, IntArray> getGraphSuccessor() {
		HashMap<Integer, IntArray> graph = new HashMap<Integer, IntArray>();
		for (int agId : this.agentIds) {
			graph.put(agId, new IntArray());
		}
		for (int agId : this.agentIds) {
			IntArray predecessors = this.getLhs(agId);
			for (int pre : predecessors) {
				graph.get(pre).add(agId);
			}
		}
		return graph;
	}

	/**
	 * 
	 * @return
	 */
	private HashMap<Integer, IntArray> getGraphPredecessor() {
		HashMap<Integer, IntArray> graph = new HashMap<Integer, IntArray>();
		for (int agId : this.agentIds) {
			IntArray predecessors = this.getLhs(agId);
			graph.put(agId, predecessors);
		}
		return graph;
	}

	public boolean isAcyclic() {
		for (int agId : this.agentIds) {
			IntArray predecessors = this.getLhs(agId);
			for (int pred : predecessors) {
				if (this.getLhs(pred).contains(agId)) {
					System.err.println("*******************************************\n" + agId +
							" :: predecessors" + predecessors + " :: pred: "
							+ pred + "\n" + this);
					return true;
				}
			}
		}
		return false;
	}

	public String toString(int agId) {
		return agId + " :: " + lhs[agId]+" --> "+this.rhs[agId];
	}

	public String toString() {
		String ans = "";
		for (int agId : this.agentIds) {
			ans += this.toString(agId) + "\n";
		}
		return ans;
	}

	public double[] computeTV(IntArray order) {
		double[] tv = new double[order.size()];
		for (int i = 0 ; i < order.size() ; i++) {
			int agId = order.get(i);
			if (this.heuristic == MinDom) {
				tv[i] = this.getRhs(agId);
			} else if (this.heuristic == DomDeg){
				tv[i] = ((double) this.rhs[agId]) / this.degrees[agId];
			} else if (this.heuristic == DomWDeg){
				tv[i] = ((double) this.rhs[agId]) / (this.weights[agId]*this.degrees[agId]);
			}
		}
		return tv;
	}
	
	/**
	 * @param agentId
	 */
	public int removeExplanationsContainning(int rhs) {
		int nbChecks = 0;
		for(int agId : this.agentIds){
			if (this.lhs[agId].get(rhs)) {
				this.removeExplanationOf(agId);
				nbChecks++;
			}
		}
		return nbChecks;
	}
	/**
	 * @param agId
	 * @param weight
	 */
	public void setWeights(int agId, int weight) {
		this.weights[agId] = weight;
	}

}
