package simulation.metric;

import java.util.Collection;
import java.util.HashMap;

import kernel.AbstractMaster;
import kernel.AgentObserver;
import kernel.StopException;
import kernel.com.Message;
import kernel.events.AgentEvent;
import util.Commands;

public class CommunicationObserver extends AgentObserver implements MetricListener {

	private HashMap<String, Metric> statisticsOfReceivedMsg;

	private long numMsgSent;

	private long numRcvdMsg;

	private long totalMsgCutOff = 5*(long) 1e7;//(long) 1e9;

	public CommunicationObserver() {
		super(Commands.MSGs);
	}

	@Override
	public void initialize() {
		super.initialize();
		this.statisticsOfReceivedMsg = new HashMap<String, Metric>();
		this.statisticsOfReceivedMsg.put(Commands.MSGs, new Metric(Commands.MSGs, 0L));
	}

	@Override
	public Collection<Metric> getEndStatistics() {
		return this.statisticsOfReceivedMsg.values();
	}

	@Override
	public boolean handleReceivedMessage(Message msg) throws StopException {
		this.numRcvdMsg++;
		if (msg.getTotatNumMessages() > this.totalMsgCutOff) {
			throw new StopException(Commands.ERROR + " Msg Cut-off"); 
		}
//		this.statisticsOfReceivedMsg.get(Commands.MSGs).value++;
		if(msg.getSenderId()!=AbstractMaster.MasterId) {
			this.statisticsOfReceivedMsg.get(Commands.MSGs).value++;
		}
		return true;
	}

	@Override
	public void handleSendMessage(Message msg) {
		this.numMsgSent++;
		if (!this.statisticsOfReceivedMsg.containsKey(msg.getType())) {
			this.statisticsOfReceivedMsg.put(msg.getType(), new Metric(msg.getType(), 1L));
		} else {
			this.statisticsOfReceivedMsg.get(msg.getType()).value++;
		}
	}

	@Override
	public void handleAgentEvent(AgentEvent event) {
	// if (event instanceof AgentFinishedEvent) {
	// Message msg = new Message(Commands.STATISTICS);
	// msg.addContents(Commands.MSGs, this.numRcvdMsg);
	// String[] handlers = this.myAgent.communicator.getMsgHandlerTypes();
	// for (String hdl : handlers) {
	// if (!this.statisticsOfReceivedMsg.containsKey(hdl)) {
	// this.statisticsOfReceivedMsg.put(hdl, 0L);
	// }
	// }
	// msg.addContents(Commands.MsgStatistics, this.statisticsOfReceivedMsg);
	// msg.setReceiverId(AbstractMaster.MasterId);
	// this.myAgent.sendMessage(msg);
	// // this.myAgent.log(Commands.nbMsgSend + " : " + this.numMsgSent,Logger.INFO_LEVEL);
	// // this.myAgent.log(Commands.nbMsgRcvd + " : " + this.numRcvdMsg,Logger.INFO_LEVEL);
	// // this.myAgent.log(Commands.MsgStatistics + " : "+ this.statisticsOfReceivedMsg.toString(), Logger.INFO_LEVEL);
	// }
	}

	@Override
	public MetricListener clone() {
		MetricListener clone = null;
		try {
			clone = (CommunicationObserver) super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return clone;
	}

	@Override
	public void handleIncrementCycle() {}

	@Override
	public void handleIncrementNbConstraintChecks(int n) {}

}
