package util.graph;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.util.LinkedList;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import gui.visualizing.TreeDrawing;
import util.IntArray;

public class PriorityTree {

	// -----------------------------------------------------------------------//
	/**
	 * root node
	 **/
	protected PriorityNode root;
	/**
	 * children of root
	 **/
	protected Vector<PriorityTree> children;

	private double pos;
	private double maxX;
	private double minX;

	public double getPosition() {
		return this.pos;
	}

	public void setPosition(double newx) {
		this.pos = newx;
	}

	public double getMinX() {
		return this.minX;
	}

	public void setMinX(double newMinX) {
		this.minX = newMinX;
	}

	public double getMaxX() {
		return this.maxX;
	}

	public void setMaxX(double newMaxX) {
		this.maxX = newMaxX;
	}

	public void position(double nextLeafX) {
		this.minX = nextLeafX;
		double firstson = nextLeafX;
		double lastson = nextLeafX;
		if (!this.children.isEmpty()) {
			boolean isfirstson = true;
			for (PriorityTree pp : this.children) {
				pp.position(nextLeafX);
				nextLeafX = pp.getMaxX() + 1;
				lastson = pp.getPosition();
				if (isfirstson) {
					firstson = pp.getPosition();
					isfirstson = false;
				}
			}
			--nextLeafX;
		}
		this.maxX = nextLeafX;
		this.pos = (firstson + lastson) / 2;
	}

	// -----------------------------------------------------------------------//
	/**
	 * Constructor
	 **/
	public PriorityTree(int Id) {
		this.root = new PriorityNode(Id, 0);
		this.children = new Vector<PriorityTree>();
	}
	/**
	 * 
	 * @param Id
	 * @param nbElements
	 */
	public PriorityTree(int Id, int nbElements) {
		root = new PriorityNode(Id, nbElements);
		children = new Vector<PriorityTree>();
	}

	public PriorityTree(PriorityNode Root) {
		this.root = Root;
		this.children = new Vector<PriorityTree>();
	}

	// -----------------------------------------------------------------------//

	@SuppressWarnings("unchecked")
	public PriorityTree clone() {
		PriorityTree clon = new PriorityTree(root.Id, root.priority);
		clon.children = (Vector<PriorityTree>) this.children.clone();
		return clon;
	}

	// -----------------------------------------------------------------------//
	public Vector<PriorityTree> getChildren() {
		return children;
	}

	// -----------------------------------------------------------------------//
	public int getRootId() {
		return root.Id;
	}

	// -----------------------------------------------------------------------//
	/**
	 * Traverse the tree until the given agent is found. Return a pointer to the subtree whose root is the parent of the given agent
	 */
	public PriorityTree getParent(int id) {
		if (this.getRootId() == id)
			return null;
		for (int i = 0 ; i < children.size() ; i++) {
			PriorityTree st = children.elementAt(i);
			if (st.root.Id == id)
				return this;
			else {
				PriorityTree sst = st.getParent(id);
				if (sst != null)
					return sst;
			}
		}
		return null;
	}

	/**
	 * Traverse the tree until the given agent is found, return its associated integer.
	 * @param id
	 * @return the the associated integer for the agent with <code>id</code>
	 */
	public int getCost(int id) {
		if (root.Id == id){
			return root.priority;
		} else{
//			for (int i = 0 ; i < children.size() ; i++) {
//				PriorityTree st = children.elementAt(i);
//				int c = st.getCost(id);
//				if (c >= 0) return c;
//			}
			for(PriorityTree st : this.children){
				int c = st.getCost(id);
				if (c >= 0) return c;
			}
		}
		return -1;
	}

	// -----------------------------------------------------------------------//
	public int size() {
		int size = 1;
		for (int i = 0 ; i < children.size() ; i++) {
			PriorityTree st = children.elementAt(i);
			size += st.size();
		}
		return size;
	}

	// -----------------------------------------------------------------------//
	/**
	 * what is the depth of this tree? 'depth' is the longest path from root to leaf
	 * */
	public int depth() {
		int maxd = 0;
		for (int i = 0 ; i < children.size() ; i++) {
			PriorityTree st = children.elementAt(i);
			int d = st.depth();
			if (d > maxd)
				maxd = d;
		}
		return maxd + 1;
	}

	// -----------------------------------------------------------------------//
	/**
	 * @return a pointer to the subtree rooted at the given agent
	 **/
	public PriorityTree getSubTree(int id) {
		if (root.Id == id)
			return this;
		for (int i = 0 ; i < children.size() ; i++) {
			PriorityTree st = children.elementAt(i);
			if (id == st.root.Id)
				return st;
			else {
				PriorityTree stt = st.getSubTree(id);
				if (stt != null)
					return stt;
			}
		}
		return null;
	}

	// -----------------------------------------------------------------------//
	/**
	 * add a branch to 'this' node.
	 **/
	public void addSubTree(PriorityTree st) {
		if (consistencyCheck(st))
			children.add(st);
		else
			throw new Error(" PriorityTree.addSubTree()\n: " + " Error in Consistency Check!!" + "\n " + this.toString() + "\n " + st.toString());
	}

	// -----------------------------------------------------------------------//
	/**
	 * is the given tree equal to this one? Equal: has the same agent/values with the same parent/child relationships. (note: costs may be different.)
	 */
	public boolean equal(PriorityTree st) {
		if (root.Id != (st.root.Id))
			return false;
		if (children.size() != st.children.size())
			return false;
		for (int i = 0 ; i < children.size() ; i++) {
			PriorityTree sst = children.elementAt(i);
			int flag = 0;
			for (int j = 0 ; j < st.children.size() ; j++) {
				PriorityTree sst2 = (PriorityTree) st.children.elementAt(j);
				if (sst.equal(sst2)) {
					flag = 1;
					break;
				}
			}
			if (flag == 0)
				return false;
		}
		return true;
	}

	/**
	 * @param id
	 */
	public IntArray getDecendants(int id) {
		IntArray res = new IntArray();
		if (this.getSubTree(id) != null) {
			res = IntArray.fromString("[" + this.getSubTree(id).toBFSString() + "]");
			res.remove(id);
		}
		return res;
	}

	public IntArray getAncestors(int id) {
		IntArray res = new IntArray();
		PriorityTree parent = this.getParent(id);
		while (parent != null) {
			res.add(parent.getRootId());
			parent = this.getParent(parent.getRootId());
		}
		return res;
	}

	/**
	 * is id in this tree?
	 * */
	public boolean isInTree(int id) {
		if (root.Id == id)
			return true;
		for (int i = 0 ; i < children.size() ; i++) {
			PriorityTree st = children.elementAt(i);
			if (st.isInTree(id))
				return true;
		}
		return false;
	}

	// -----------------------------------------------------------------------//
	/**
	 * another consistency check method is id1 a descendent of id2?
	 * */
	public boolean isDescendent(int id1, int id2) {
		if (root.Id == id1)
			for (int i = 0 ; i < children.size() ; i++) {
				PriorityTree st = children.elementAt(i);
				if (st.isInTree(id2))
					return true;
			}
		else
			for (int i = 0 ; i < children.size() ; i++) {
				PriorityTree st = children.elementAt(i);
				if (st.isDescendent(id1, id2))
					return true;
			}
		return false;
	}

	// -----------------------------------------------------------------------//
	/**
	 * is any agent in st already in this tree?
	 **/
	public boolean consistencyCheck(PriorityTree st) {
		if (this.isInTree(st.root.Id))
			return false;
		for (int i = 0 ; i < st.children.size() ; i++) {
			PriorityTree sst = (PriorityTree) st.children.elementAt(i);
			if (!consistencyCheck(sst)) { return false; }
		}
		return true;
	}

	/**
	 * Returns a string representation of this object
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("{" + this.root.toString());
		for (int i = 0 ; i < children.size() ; i++) {
			PriorityTree st = children.elementAt(i);
			sb.append(", " + st.toString());
		}
		sb.append("}");
		return sb.toString();
	}

	// -----------------------------------------------------------------------//

	public String toBFSString() {
		StringBuffer sb = new StringBuffer();
		LinkedList<PriorityTree> q = new LinkedList<PriorityTree>();
		q.addLast(this);
		sb.append(root.Id);
		while (!q.isEmpty()) {
			PriorityTree pt = q.removeFirst();
			for (int i = 0 ; i < pt.children.size() ; i++) {
				PriorityTree st = (PriorityTree) pt.children.elementAt(i);
				sb.append(" " + st.root.Id);
				q.addLast(st);
			}
		}
		return sb.toString();
	}

	// -----------------------------------------------------------------------//

	public String toDFSString() {
		StringBuffer sb = new StringBuffer();
		sb.append(root.Id);
		for (PriorityTree st : children) {
			sb.append(" " + st.toDFSString());
		}
		return sb.toString();
	}

	// -----------------------------------------------------------------------//

	public static PriorityTree fromString(String str) {
		PriorityTree result = null;
		char[] s = str.toCharArray();
		int i = 0;
		while (s[i++] != '{');
		i--;
		int j = findClosingBrace(s, i);
		result = parseHelper(s, i, j);
		return result;
	}

	// -----------------------------------------------------------------------//
	/**
	 * s[idx] should be an open brace. returns i such that s[i] is the matching closing brace
	 */
	public static int findClosingBrace(char[] s, int idx) {
		int numopen = 1;
		int numclosed = 0;
		int i = idx + 1;
		while (i < s.length) {
			char c = s[i++];
			if (c == '}') {
				numclosed++;
				if (numopen == numclosed)
					return i - 1;
			}
			if (c == '{')
				numopen++;
		}
		return i;
	}

	// -----------------------------------------------------------------------//
	/**
	 * return the PriorityTree from s[idx] to s[idx2] s[idx] should be the opening brace of a PriorityTree. s[idx2] should be the closing brace of a
	 * PriorityTree.
	 */
	public static PriorityTree parseHelper(char[] s, int idx, int idx2) {
		PriorityTree result = null;
		int i = idx;
		int j = idx2;
		if (s[i++] != '{') { /* start of PriorityTree should be { */
			System.out.println("PriorityTree.parseString(): Parse error: start of Prioritytree should be \"{\"");
			System.exit(0);
		} else {
			if (s[i++] != '(') { /* start of root node should be ( */
				System.out.println("PriorityTree.parseString(): Parse error: start of root node should be (!!");
				System.exit(0);
			} else {
				char[] ss = new char[100];
				int ii = 0;
				while (s[i] != ')') {
					ss[ii] = s[i];
					ii++;
					i++;
				}
				ss[ii] = '\0';
				PriorityNode root = PriorityNode.fromString(new String(ss));
				result = new PriorityTree(root);
				i++;
				if (s[i] == '}') /* we are done */
					return result;
				else {
					while (s[i++] != '{');/* get to next open brace */
					i--;
					while (i < j) {
						int jj = PriorityTree.findClosingBrace(s, i);
						PriorityTree st = PriorityTree.parseHelper(s, i, jj);
						result.addSubTree(st);
						i = jj;
						i++;
						if (s[i] == '}') /* we are done */
							return result;
						else {
							while (s[i++] != '{'); /* get to next open brace */
							i--;
						}
					}
				}
			}
		}
		return result;
	}

	public String pretty() {
		StringBuffer sb = new StringBuffer();
		sb.append("{A" + this.root.getNodeId());
		for (int i = 0 ; i < children.size() ; i++) {
			PriorityTree st = children.elementAt(i);
			sb.append("" + st.pretty());
		}
		sb.append("}");
		return sb.toString();
	}

	public String toDrawableTree() {
		StringBuffer sb = new StringBuffer();
		sb.append("{A" + this.root.Id);
		for (int i = 0 ; i < children.size() ; i++) {
			PriorityTree st = children.elementAt(i);
			sb.append("" + st.toDrawableTree());
		}
		sb.append("}");
		return sb.toString();
	}

	// -----------------------------------------------------------------------//

	public static void main(String args[]) {
		final String s //= "{(1,0), {(2,4), {( 4, 5), {(5, 7)}}}, {(3,4), {( 7, 5), {(8, 7)}}}}";
		//= "{(7,48), {(19,47), {(22,46), {(28,45), {(31,44), {(37,43), {(43,42), {(44,41), {(14,40), {(13,39), {(1,38), {(10,37), {(16,36), {(17,35), {(23,34), {(26,33), {(25,32), {(27,31), {(4,30), {(8,29), {(15,28), {(39,27), {(38,26)}, {(41,26), {(40,25), {(20,24), {(29,23), {(34,22), {(32,21), {(47,20), {(46,19), {(48,18), {(6,17), {(11,16), {(21,15), {(33,14)}}, {(12,15), {(18,14), {(24,13), {(42,12)}}}}}, {(5,16), {(2,15), {(3,14), {(9,13), {(36,12), {(35,11), {(30,10)}}, {(45,11)}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}";
		//= "{(12,20), {(17,19), {(16,18), {(6,17), {(19,16), {(5,15), {(7,14), {(2,13)}}, {(9,14), {(10,13), {(1,12)}}, {(20,13), {(3,12), {(4,11), {(8,10)}, {(18,10), {(14,9)}}}, {(15,11)}}}}}}}, {(13,17)}}, {(11,18)}}}";
		= "{(19,20), {(8,19), {(4,18), {(9,17), {(10,16), {(7,15), {(17,14), {(15,13), {(18,12), {(2,11), {(1,10), {(11,9), {(14,8), {(13,7), {(20,6), {(3,5), {(12,4), {(5,3), {(16,2)}}}}}}, {(6,7)}}}}}}}}}}}}}}";
		for (UIManager.LookAndFeelInfo laf : UIManager.getInstalledLookAndFeels()) {
			if ("Mac OS X".equals(laf.getName())) {
				try {
					UIManager.setLookAndFeel(laf.getClassName());
				} catch (Exception e) {}
			} else {
				if ("Nimbus".equals(laf.getName())) {
					try {
						UIManager.setLookAndFeel(laf.getClassName());
					} catch (Exception e) {}
				}
			}
		}
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				JFrame frame = new JFrame();
				frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
				frame.setTitle("DisChoco: Distributed CSP Problem Simulator");
				frame.setMinimumSize(new Dimension(1024, 720));
				frame.setMaximumSize(new Dimension(1024, 720));
				frame.setPreferredSize(new Dimension(1024, 720));
				Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
				// Determine the new location of the window
				int w = frame.getSize().width;
				int h = frame.getSize().height;
				int x = (dim.width - w) / 2;
				int y = (dim.height - h) / 2;
				// Move the window
				frame.setLocation(x, y);
				PriorityTree pt = PriorityTree.fromString(s);
				TreeDrawing treePanel = new TreeDrawing(pt, new Dimension(1024, 720));
				frame.getContentPane().add(treePanel, BorderLayout.CENTER);
				frame.pack();
				frame.setVisible(true);
			}
		});
//		PriorityTree pt = PriorityTree.fromString(s);
//		final TreeDrawing treePanel = new TreeDrawing(pt, new Dimension(1024, 720));
//		System.out.println("String " + s);
//		System.out.println("pretty " + pt.pretty());
//		System.out.println("toString " + pt.toString());
//		System.out.println("toDFSString " + pt.toDFSString());
//		System.out.println("toBFSString " + pt.toBFSString());
//		System.out.println("isDescendent(3,1) " + pt.isDescendent(1, 3));
	}

}
